#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010-2011 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#
#   uvm-python NOTE: All code ported from SystemVerilog UVM 1.2 to
#   python. Original code structures (including comments)
#   preserved where possible.
#----------------------------------------------------------------------

from typing import Dict, List, Optional

from .sv import sv
from .uvm_object import UVMObject
from .uvm_report_object import UVMReportObject
from .uvm_pool import UVMPool
from .uvm_queue import UVMQueue
from .uvm_misc import UVM_APPEND
from ..macros import (uvm_cb_trace_noobj, uvm_warning, uvm_info)
from ..uvm_macros import UVM_STRING_QUEUE_STREAMING_PACK
from .uvm_object_globals import UVM_NONE
from .uvm_globals import (uvm_report_warning, uvm_report_error,
    uvm_report_fatal)

ALL_TYPES = None

"""
Title: Callbacks Classes

This section defines the classes used for callback registration, management,
and user-defined callbacks.
"""


def o2str(ordering):
    if ordering == UVM_APPEND:
        return "UVM_APPEND"
    return "<NO_ORDER_ADDED>"



class UVMTypeIDBase:
    """
    Simple typeid interface. Need this to set up the base-super mapping.
    This is similar to the factory, but much simpler. The idea of this
    interface is that each object type T has a typeid that can be
    used for mapping type relationships. This is not a user visible class.
    """

    typename = ""

    # UVMTypeIDBase -> uvm_callbacks
    typeid_map: Dict['UVMTypeIDBase', 'UVMCallbacks'] = {}

    # uvm_callbacks_base -> UVMTypeIDBase
    type_map: Dict['UVMCallbacksBase', 'UVMTypeIDBase'] = {}


class UVMTypeID(UVMTypeIDBase):
    m_b_inst = None

    @classmethod
    def get(cls):
        if UVMTypeID.m_b_inst is None:
            UVMTypeID.m_b_inst = UVMTypeID()
        return UVMTypeID.m_b_inst


class UVMCallbacksBase(UVMObject):
    """
    Base class singleton that holds generic queues for all instance
    specific objects. This is an internal class. This class contains a
    global pool that has all of the instance specific callback queues in it.
    All of the typewide callback queues live in the derivative class
    UVMTypedCallbacks#(T). This is not a user visible class.

    This class holds the class inheritance hierarchy information
    (super types and derivative types).

    Note, all derivative uvm_callbacks#() class singletons access this
    global m_pool object in order to get access to their specific
    instance queue.
    """

    m_b_inst = None
    m_pool = UVMPool()  # uvm_object -> uvm_queue#(uvm_callback)
    m_tracing = True

    def __init__(self, name):
        super().__init__(name)
        #Type checking interface
        self.m_this_type = []     # one to many T->T/CB
        self.m_super_type = None       # one to one relation
        self.m_derived_types = []  # one to many relation, UVMTypeIDBase

    @classmethod
    def m_initialize(cls):
        if cls.m_b_inst is None:
            cls.m_b_inst = UVMCallbacksBase("callbacks_base")
            cls.m_pool = UVMPool()
        return UVMCallbacksBase.m_b_inst

    def m_am_i_a(self, obj):
        return 0

    def m_is_for_me(self, cb: 'UVMCallback'):
        return 0

    def m_is_registered(self, obj, cb):
        return True

    def m_get_tw_cb_q(self, obj, CB=None):
        return None

    def m_add_tw_cbs(self, cb, ordering):
        raise Exception('m_add_tw_cbs is pure virtual function!')

    def m_delete_tw_cbs(self, cb):
        return 0

    def check_registration(self, obj, cb):
        """
        Check registration. To test registration, start at this class and
        work down the class hierarchy. If any class returns true then
        the pair is legal.

        Args:
            obj (UVMObject):
            cb (UVMCallback):

        Returns:
            bool: True if obj is registered, False otherwise
        """
        # st = None
        dt = None

        if self.m_is_registered(obj,cb):
            return True

        # Need to look at all possible T/CB pairs of this type
        for item in self.m_this_type:
            if (UVMCallbacksBase.m_b_inst != item and
                    item.m_is_registered(obj,cb)):
                return True

        if obj is None:
            for item in self.m_derived_types:
                dt = UVMTypeIDBase.typeid_map[item]
                if dt is not None and dt.check_registration(None,cb):
                    return True

        return False

    @classmethod
    def get_first(cls, itr, obj, CB=None) -> Optional['UVMCallback']:
        """
        Returns the first enabled callback of type CB which resides in the queue for `obj`.
        If `obj` is `None` then the typewide queue for T is searched. `itr` is the iterator
        it will be updated with a value that can be supplied to `get_next` to get the next
        callback object.

        If the queue is empty then `None` is returned.

        The iterator class `uvm_callback_iter` may be used as an alternative, simplified,
        iterator interface.

        Args:
            itr:
            obj:
            CB: Type of the callback
        Returns:
            UVMCallback: First found callback.
        """
        cb = None
        cls.get()
        q = cls.m_get_q(obj, CB)
        if q is not None:
            for i in range(len(q)):
                itr.m_i = i
                cb = q[i]
                if cb is not None and cb.callback_mode():
                    return cb
        return None


#------------------------------------------------------------------------------
#
# Class - UVMTypedCallbacks#(T)
#
#------------------------------------------------------------------------------

# Another internal class. This contains the queue of typewide
# callbacks. It also contains some of the public interface methods,
# but those methods are accessed via the uvm_callbacks#() class
# so they are documented in that class even though the implementation
# is in this class.
#
# The <add>, <delete>, and <display> methods are implemented in this class.


class UVMTypedCallbacks(UVMCallbacksBase):  # (type T=uvm_object) extends uvm_callbacks_base

    #  static uvm_queue#(uvm_callback) m_tw_cb_q
    m_tw_cb_q: List['UVMCallback'] = []

    #  static string m_typename
    m_typename = ""


    #  #The actual global object from the derivative class. Note that this is
    #  #just a reference to the object that is generated in the derived class.
    #  static this_type m_t_inst
    m_t_inst = None

    def __init__(self, name, T=ALL_TYPES):
        super().__init__(name)
        self.T = T

    @classmethod
    def m_initialize(cls):
        if UVMTypedCallbacks.m_t_inst is None:
            super().m_initialize()
            UVMTypedCallbacks.m_t_inst = UVMTypedCallbacks("typed_callbacks")
            UVMTypedCallbacks.m_t_inst.m_tw_cb_q = []
        return UVMTypedCallbacks.m_t_inst


    def m_am_i_a(self, obj):
        """
        Type checking interface: is given `obj` of type T?
        Args:
            obj:
        Returns:
        """
        if obj is None:
            return True
        arr = []
        if self.T is not ALL_TYPES:
            return sv.cast(arr, obj, self.T)
        return True


    def m_get_tw_cb_q(self, obj, CB=None):
        """
        Getting the typewide queue
        Args:
            obj:
            CB:
        Returns:
        """
        if self.m_am_i_a(obj):
            for i in range(len(self.m_derived_types)):
                #super_type dt
                dt = UVMTypeIDBase.typeid_map[self.m_derived_types[i]]
                if dt is not None and dt != self:
                    m_get_tw_cb_q = dt.m_get_tw_cb_q(obj, CB)
                    if m_get_tw_cb_q is not None:
                        return m_get_tw_cb_q
            return UVMTypedCallbacks.m_t_inst.m_tw_cb_q
        else:
            return None


    @classmethod
    def m_cb_find(cls, q, cb):
        for i in range(len(q)):
            if q[i] == cb:
                return i
        return -1


    @classmethod
    def m_cb_find_name(cls, q, name, where):
        """
         static function int m_cb_find_name(uvm_queue#(uvm_callback) q, string name, string where)
        Args:
            cls:
            q:
            name:
            where:
        Returns:
        """
        cb = None
        for i in range(len(q)):
            cb = q[i]
            if cb.get_name() == name:
                uvm_report_warning("UVM/CB/NAM/SAM", "A callback named \"" + name
                    + "\" is already registered with " + where)
                return 1
        return 0
        #  endfunction


    def m_add_tw_cbs(self, cb, ordering):
        """
         #For a typewide callback, need to add to derivative types as well.
         virtual function void m_add_tw_cbs(uvm_callback cb, uvm_apprepend ordering)
        Args:
            cb:
            ordering:
        """
        cb_pair = None  # super_type
        obj = None  # uvm_object
        me = None  # T
        warned = False

        # tpoikela: Which one is correct?
        #cls = UVMCallbacks
        cls = UVMTypedCallbacks

        q = []  # uvm_queue#(uvm_callback)
        if self.m_cb_find(cls.m_t_inst.m_tw_cb_q, cb) == -1:
            warned = cls.m_cb_find_name(cls.m_t_inst.m_tw_cb_q, cb.get_name(), "type")
            if ordering == UVM_APPEND:
                cls.m_t_inst.m_tw_cb_q.append(cb)
            else:
                cls.m_t_inst.m_tw_cb_q.insert(0, cb)

        if cls.m_t_inst.m_pool.has_first():
            obj = cls.m_t_inst.m_pool.get_first()
            while True:
                #if($cast(me,obj)):
                me = obj
                q = cls.m_t_inst.m_pool.get(obj)
                if q is None:
                    q = []
                    cls.m_t_inst.m_pool.add(obj,q)

                if self.m_cb_find(q,cb) == -1:
                    if not warned:
                        cls.m_cb_find_name(q, cb.get_name(), "object instance "
                                + me.get_full_name())

                    if ordering == UVM_APPEND:
                        q.append(cb)
                    else:
                        q.insert(0, cb)

                if cls.m_t_inst.m_pool.has_next() is False:
                    break
                #end while(m_t_inst.m_pool.next(obj))

        for i in range(len(self.m_derived_types)):
            cb_pair = UVMTypeIDBase.typeid_map[self.m_derived_types[i]]
            if cb_pair != self:
                cb_pair.m_add_tw_cbs(cb,ordering)

    #  endfunction


    #  #For a typewide callback, need to remove from derivative types as well.
    #  virtual function bit m_delete_tw_cbs(uvm_callback cb)
    #    super_type cb_pair
    #    uvm_object obj
    #    uvm_queue#(uvm_callback) q
    #    int pos = m_cb_find(m_t_inst.m_tw_cb_q,cb)
    #
    #    if(pos != -1):
    #      m_t_inst.m_tw_cb_q.delete(pos)
    #      m_delete_tw_cbs = 1
    #    end
    #
    #    if(m_t_inst.m_pool.first(obj)):
    #      do begin
    #        q = m_t_inst.m_pool.get(obj)
    #        if(qis None):
    #          q=new
    #          m_t_inst.m_pool.add(obj,q)
    #        end
    #        pos = m_cb_find(q,cb)
    #        if(pos != -1):
    #          q.delete(pos)
    #          m_delete_tw_cbs = 1
    #        end
    #      end while(m_t_inst.m_pool.next(obj))
    #    end
    #    foreach(m_derived_types[i]):
    #      cb_pair = UVMTypeIDBase::typeid_map[m_derived_types[i] ]
    #      if(cb_pair != this)
    #        m_delete_tw_cbs |= cb_pair.m_delete_tw_cbs(cb)
    #    end
    #  endfunction


    @classmethod
    def display(cls, obj=None):
        """
         static function void display(T obj=None)
        Args:
            cls:
            obj:
        """
        #    T me
        #    super_type ib = m_t_inst
        cbq = []  # string [$]
        inst_q = []  # string [$]
        mode_q = []  # string  [$]
        cb = None  # uvm_callback
        blanks = "                             "
        bobj = obj
        qs = []  # string[$]

        q = None  # uvm_queue#(uvm_callback)
        tname = ""
        strr = ""
        max_cb_name = 0
        max_inst_name = 0

        prev_trace = cls.m_tracing
        cls.m_tracing = False  # don't allow tracing during display

        if cls.m_typename != "":
            tname = cls.m_typename
        elif (obj is not None):
            tname = obj.get_type_name()
        else:
            tname = "*"

        q = UVMTypedCallbacks.m_t_inst.m_tw_cb_q
        for i in range(len(q)):
            cb = q[i]
            cbq.append(cb.get_name())
            inst_q.append("(*)")
            if (cb.is_enabled()):
                mode_q.append("ON")
            else:
                mode_q.append("OFF")

            strr = cb.get_name()
            if max_cb_name <= len(strr):
                max_cb_name = len(strr)
            strr = "(*)"
            if max_inst_name <= len(strr):
                max_inst_name = len(strr)

        # tpoikela: TODO finalize this
        #    if(obj is None):
        #      if(m_t_inst.m_pool.first(bobj)):
        #        do
        #          if($cast(me,bobj)) break
        #        while(m_t_inst.m_pool.next(bobj))
        #      end
        #      if(me is not None || m_t_inst.m_tw_cb_q.size()):
        #        qs.push_back($sformatf("Registered callbacks for all instances of %s\n", tname))
        #        qs.push_back("---------------------------------------------------------------\n")
        #      end
        #      if(me is not None):
        #        do begin
        #          if($cast(me,bobj)):
        #            q = m_t_inst.m_pool.get(bobj)
        #            if (qis None):
        #              q=new
        #              m_t_inst.m_pool.add(bobj,q)
        #            end
        #            for(int i=0; i<q.size(); ++i):
        #              cb = q.get(i)
        #              cbq.push_back(cb.get_name())
        #              inst_q.push_back(bobj.get_full_name())
        #              if(cb.is_enabled()) mode_q.push_back("ON")
        #              else mode_q.push_back("OFF")
        #
        #              strr = cb.get_name()
        #              max_cb_name = max_cb_name > strr.len() ? max_cb_name : strr.len()
        #              strr = bobj.get_full_name()
        #              max_inst_name = max_inst_name > strr.len() ? max_inst_name : strr.len()
        #            end
        #          end
        #        end while (m_t_inst.m_pool.next(bobj))
        #      end
        #      else begin
        #        qs.push_back($sformatf("No callbacks registered for any instances of type %s\n", tname))
        #      end
        #    end
        #    else begin
        #      if(m_t_inst.m_pool.exists(bobj) || m_t_inst.m_tw_cb_q.size()):
        #       qs.push_back($sformatf("Registered callbacks for instance %s of %s\n", obj.get_full_name(), tname))
        #       qs.push_back("---------------------------------------------------------------\n")
        #      end
        #      if(m_t_inst.m_pool.exists(bobj)):
        #        q = m_t_inst.m_pool.get(bobj)
        #        if(qis None):
        #          q=new
        #          m_t_inst.m_pool.add(bobj,q)
        #        end
        #        for(int i=0; i<q.size(); ++i):
        #          cb = q.get(i)
        #          cbq.push_back(cb.get_name())
        #          inst_q.push_back(bobj.get_full_name())
        #          if(cb.is_enabled()) mode_q.push_back("ON")
        #          else mode_q.push_back("OFF")
        #
        #          str = cb.get_name()
        #          max_cb_name = max_cb_name > str.len() ? max_cb_name : str.len()
        #          str = bobj.get_full_name()
        #          max_inst_name = max_inst_name > str.len() ? max_inst_name : str.len()
        #        end
        #      end
        #    end
        #    if(!cbq.size()):
        #      if(obj is None) str = "*"
        #      else str = obj.get_full_name()
        #      qs.push_back($sformatf("No callbacks registered for instance %s of type %s\n", str, tname))
        #    end
        #
        for i in range(len(cbq)):
            padding = blanks[0:max_cb_name - len(cbq[i])]
            qs.append(sv.sformatf("%s  %s %s on %s  %s\n",
                cbq[i], padding, inst_q[i], blanks[0:max_inst_name - len(inst_q[i])-1],
                mode_q[i]))
        uvm_info("UVM/CB/DISPLAY", UVM_STRING_QUEUE_STREAMING_PACK(qs), UVM_NONE)
        cls.m_tracing = prev_trace  # allow tracing to be resumed



class UVMCallbacks(UVMTypedCallbacks):
    """
    The ~uvm_callbacks~ class provides a base class for implementing callbacks,
    which are typically used to modify or augment component behavior without
    changing the component class. To work effectively, the developer of the
    component class defines a set of "hook" methods that enable users to
    customize certain behaviors of the component in a manner that is controlled
    by the component developer. The integrity of the component's overall behavior
    is intact, while still allowing certain customizable actions by the user.

    To enable compile-time type-safety, the class is parameterized on both the
    user-defined callback interface implementation as well as the object type
    associated with the callback. The object type-callback type pair are
    associated together using the `UVMRegisterCb` macro to define
    a valid pairing; valid pairings are checked when a user attempts to add
    a callback to an object.

    To provide the most flexibility for end-user customization and reuse, it
    is recommended that the component developer also define a corresponding set
    of virtual method hooks in the component itself. This affords users the ability
    to customize via inheritance/factory overrides as well as callback object
    registration. The implementation of each virtual method would provide the
    default traversal algorithm for the particular callback being called. Being
    virtual, users can define subtypes that override the default algorithm,
    perform tasks before and/or after calling super.<method> to execute any
    registered callbacks, or to not call the base implementation, effectively
    disabling that particular hook. A demonstration of this methodology is
    provided in an example included in the kit.
    """

    # Singleton instance is used for type checking
    m_inst = None
    # typeinfo
    m_typeid = None  # type: UVMTypeIDBase
    m_cb_typeid = None  # type: UVMTypeIDBase

    m_typename = ''
    m_cb_typename = ''
    reporter = UVMReportObject("cb_tracer")
    m_base_inst = None  # uvm_callbacks#(T,uvm_callback)

    # tpoikela: Added for containing callbacks for each class
    _m_cb_table: Dict[str, 'UVMCallbacks'] = {}  # UVMCallbacks[str]

    def __init__(self, name='uvm_callbacks', T=ALL_TYPES, CB=ALL_TYPES):
        super().__init__(name)
        self.m_registered = False
        # These are type parameters
        self.T = T
        self.CB = CB


    @classmethod
    def get(cls):
        """
         get
         ---

        Returns:
        """

        if cls.m_inst is None:
            cb_base_type = UVMTypeIDBase()
            super().m_initialize()

            cb_base_type = UVMTypeID.get()
            cls.m_cb_typeid  = UVMTypeID.get()
            cls.m_typeid     = UVMTypeID.get()

            cls.m_inst = UVMCallbacks()

            if cb_base_type == cls.m_cb_typeid:
                #cast(m_base_inst, m_inst)
                cls.m_base_inst = cls.m_inst
                # The base inst in the super class gets set to this base inst
                cls.m_t_inst = cls.m_base_inst
                UVMTypeIDBase.typeid_map[cls.m_typeid] = cls.m_inst
                UVMTypeIDBase.type_map[cls.m_b_inst] = cls.m_typeid
            else:
                cls.m_base_inst = cls.get()
                cls.m_base_inst.m_this_type.append(cls.m_inst)

            if cls.m_inst is None:
                uvm_report_fatal("CB/INTERNAL","get(): m_inst is None")

        return cls.m_inst


    @classmethod
    def m_register_pair(cls, tname="", cbname="", T=ALL_TYPES):
        """
         m_register_pair
         -------------
         Register valid callback type

        Args:
            cls:
            tname:
            cbname:
            T:
            ALL_TYPES:
        Returns:
        """
        inst = cls.get()
        # tpoikela: mimics typed callbacks by creating one cbs-object
        # per registered pair
        cbs = UVMCallbacks(tname + "__" + cbname, T)

        cbs.m_typename = tname
        # TODO super_type.m_typename = tname
        cbs.m_typeid.typename = tname

        cbs.m_cb_typename = cbname
        cbs.m_cb_typeid.typename = cbname

        cbs.m_registered = True
        if cbs.get_name() not in inst._m_cb_table:
            inst._m_cb_table[cbs.get_name()] = cbs
        else:
            uvm_warning("CB_EXISTS", "Callback for " + cbs.get_name() + " already register")
        return True

    @classmethod
    def _get_typed_cbs(cls, obj, CB):
        cb_table_key = cls._get_cb_table_key(obj, CB)
        if cb_table_key in cls._m_cb_table:
            return cls._m_cb_table[cb_table_key]
        return cls

    @classmethod
    def _get_cb_table_key(cls, obj, CB):
        if obj is not None:
            if CB is not None:
                return type(obj).__name__ + '__' + CB.__name__
            else:
                return type(obj).__name__
        elif CB is not None:
            return '__' + CB.__name__
        else:
            return '__<unknown_key>__'


    #  virtual function bit m_is_registered(uvm_object obj, uvm_callback cb)
    #    if(m_is_for_me(cb) && m_am_i_a(obj)):
    #      return m_registered
    #    end
    #  endfunction

    #  #Does type check to see if the callback is valid for this type
    #  virtual function bit m_is_for_me(uvm_callback cb)
    #    CB this_cb
    #    return($cast(this_cb,cb))
    #  endfunction

    # Group: Add/delete interface

    # Function: add
    #
    # Registers the given callback object, ~cb~, with the given
    # ~obj~ handle. The ~obj~ handle can be ~None~, which allows
    # registration of callbacks without an object context. If
    # ~ordering~ is UVM_APPEND (default), the callback will be executed
    # after previously added callbacks, else  the callback
    # will be executed ahead of previously added callbacks. The ~cb~
    # is the callback handle; it must be non-~None~, and if the callback
    # has already been added to the object instance then a warning is
    # issued. Note that the CB parameter is optional. For example, the
    # following are equivalent:
    #
    #| uvm_callbacks#(my_comp)::add(comp_a, cb)
    #| uvm_callbacks#(my_comp, my_callback)::add(comp_a,cb)
    @classmethod
    def add(cls, obj, cb, ordering=UVM_APPEND):
        q = []  # uvm_queue#(uvm_callback) q
        nm = ""
        tnm = ""
        cls.get()

        if cb is None:
            nm, tnm = cls.get_obj_and_typename(obj)

            uvm_report_error("CBUNREG",
                "None callback object cannot be registered with object "
                     + nm + " (" + tnm + ")", UVM_NONE)
            return


        if not cls.m_base_inst.check_registration(obj,cb):
            nm, tnm = cls.get_obj_and_typename(obj)

            uvm_report_warning("CBUNREG",
                "Callback " + cb.get_name() + " cannot be registered with object "
                + nm + " because callback type " + cb.get_type_name()
                + " is not registered with object type " + tnm, UVM_NONE)


        if obj is None:
            if cls.m_cb_find(cls.m_t_inst.m_tw_cb_q, cb) != -1:

                if cls.m_base_inst.m_typename != "":
                    tnm = cls.m_base_inst.m_typename
                else:
                    tnm = "uvm_object"

                uvm_report_warning("CBPREG",
                    "Callback object " + cb.get_name()
                    + " is already registered with type " + tnm, UVM_NONE)
            else:
                uvm_cb_trace_noobj(cb,sv.sformatf("Add (%s) typewide callback %0s for type %s",
                    o2str(ordering), cb.get_name(), cls.m_base_inst.m_typename))
                cls.m_t_inst.m_add_tw_cbs(cb, ordering)

        else:

            uvm_cb_trace_noobj(cb, sv.sformatf("Add (%s) callback %0s to object %0s ",
                o2str(ordering), cb.get_name(), obj.get_full_name()))

            q = cls.m_base_inst.m_pool.get(obj)

            if q is None:
                q = []
                cls.m_base_inst.m_pool.add(obj,q)

            if len(q) == 0:
                # Need to make sure that registered report catchers are added. This
                # way users don't need to set up uvm_report_object as a super type.
                o = []  # uvm_report_object

                if (sv.cast(o,obj, UVMReportObject)):
                    # uvm_queue#(uvm_callback) qr
                    UVMCallbacks.get()
                    qr = UVMCallbacks.m_t_inst.m_tw_cb_q
                    for i in range(len(qr)):
                        q.append(qr[i])

                # for(int i=0; i<m_t_inst.m_tw_cb_q.size(); ++i)
                for i in range(len(cls.m_t_inst.m_tw_cb_q)):
                    q.append(cls.m_t_inst.m_tw_cb_q[i])


            #check if already exists in the queue
            if cls.m_cb_find(q,cb) != -1:
                uvm_report_warning("CBPREG", "Callback object " + cb.get_name()
                    + " is already registered with object " + obj.get_full_name(), UVM_NONE)
            else:
                cls.m_cb_find_name(q, cb.get_name(), "object instance " + obj.get_full_name())
                if ordering == UVM_APPEND:
                    q.append(cb)
                else:
                    q.insert(0, cb)

        #  endfunction

    @classmethod
    def get_obj_and_typename(cls, obj):
        nm = ""
        tnm = ""

        if obj is None:
            nm = "(*)"
        else:
            nm = obj.get_full_name()

        if cls.m_base_inst.m_typename != "":
            tnm = cls.m_base_inst.m_typename
        elif obj is not None:
            tnm = obj.get_type_name()
        else:
            tnm = "uvm_object"
        return nm, tnm


    #  # Function: add_by_name
    #  #
    #  # Registers the given callback object, ~cb~, with one or more uvm_components.
    #  # The components must already exist and must be type T or a derivative. As
    #  # with <add> the CB parameter is optional. ~root~ specifies the location in
    #  # the component hierarchy to start the search for ~name~. See <uvm_root::find_all>
    #  # for more details on searching by name.
    #
    #  static function void add_by_name(string name,
    #                                   uvm_callback cb,
    #                                   uvm_component root,
    #                                   uvm_apprepend ordering=UVM_APPEND)
    #    uvm_component cq[$]
    #    uvm_root top
    #    uvm_coreservice_t cs
    #    T t
    #    void'(get())
    #    cs = uvm_coreservice_t::get()
    #    top = cs.get_root()
    #
    #    if(cbis None):
    #       uvm_report_error("CBUNREG", { "None callback object cannot be registered with object(s) ",
    #         name }, UVM_NONE)
    #       return
    #    end
    #    `uvm_cb_trace_noobj(cb,$sformatf("Add (%s) callback %0s by name to object(s) %0s ",
    #                    ordering.name(), cb.get_name(), name))
    #    top.find_all(name,cq,root)
    #    if(cq.size() == 0):
    #      uvm_report_warning("CBNOMTC", { "add_by_name failed to find any components matching the name ",
    #        name, ", callback ", cb.get_name(), " will not be registered." }, UVM_NONE)
    #    end
    #    foreach(cq[i]):
    #      if($cast(t,cq[i])):
    #        add(t,cb,ordering)
    #      end
    #    end
    #  endfunction
    #
    #
    #  # Function: delete
    #  #
    #  # Deletes the given callback object, ~cb~, from the queue associated with
    #  #  the given ~obj~ handle. The ~obj~ handle can be ~None~, which allows
    #  # de-registration of callbacks without an object context.
    #  # The ~cb~ is the callback handle; it must be non-~None~, and if the callback
    #  # has already been removed from the object instance then a warning is
    #  # issued. Note that the CB parameter is optional. For example, the
    #  # following are equivalent:
    #  #
    #  #| uvm_callbacks#(my_comp)::delete(comp_a, cb)
    #  #| uvm_callbacks#(my_comp, my_callback)::delete(comp_a,cb)
    #
    #  static function void delete(T obj, uvm_callback cb)
    #    uvm_object b_obj = obj
    #    uvm_queue#(uvm_callback) q
    #    bit found
    #    int pos
    #    void'(get())
    #
    #    if(obj is None):
    #      `uvm_cb_trace_noobj(cb,$sformatf("Delete typewide callback %0s for type %s",
    #                       cb.get_name(), m_base_inst.m_typename))
    #      found = m_t_inst.m_delete_tw_cbs(cb)
    #    end
    #    else begin
    #      `uvm_cb_trace_noobj(cb,$sformatf("Delete callback %0s from object %0s ",
    #                      cb.get_name(), obj.get_full_name()))
    #      q = m_base_inst.m_pool.get(b_obj)
    #      pos = m_cb_find(q,cb)
    #      if(pos != -1):
    #        q.delete(pos)
    #        found = 1
    #      end
    #    end
    #    if(!found):
    #      string nm
    #      if(objis None) nm = "(*)"; else nm = obj.get_full_name()
    #      uvm_report_warning("CBUNREG", { "Callback ", cb.get_name(), " cannot be removed from object ",
    #        nm, " because it is not currently registered to that object." }, UVM_NONE)
    #    end
    #  endfunction
    #
    #

    #  # Function: delete_by_name
    #  #
    #  # Removes the given callback object, ~cb~, associated with one or more
    #  # uvm_component callback queues. As with <delete> the CB parameter is
    #  # optional. ~root~ specifies the location in the component hierarchy to start
    #  # the search for ~name~. See <uvm_root::find_all> for more details on searching
    #  # by name.
    #
    #  static function void delete_by_name(string name, uvm_callback cb,
    #     uvm_component root)
    #    uvm_component cq[$]
    #    uvm_root top
    #    T t
    #    uvm_coreservice_t cs
    #    void'(get())
    #    cs = uvm_coreservice_t::get()
    #    top = cs.get_root()
    #
    #    `uvm_cb_trace_noobj(cb,$sformatf("Delete callback %0s by name from object(s) %0s ",
    #                    cb.get_name(), name))
    #    top.find_all(name,cq,root)
    #    if(cq.size() == 0):
    #      uvm_report_warning("CBNOMTC", { "delete_by_name failed to find any components matching the name ",
    #        name, ", callback ", cb.get_name(), " will not be unregistered." }, UVM_NONE)
    #    end
    #    foreach(cq[i]):
    #      if($cast(t,cq[i])):
    #        delete(t,cb)
    #      end
    #    end
    #  endfunction


    @classmethod
    def m_get_q(cls, obj, CB=None):
        """
        #--------------------------
        Group: Iterator Interface
        #--------------------------

        This set of functions provide an iterator interface for callback queues.
        A facade class, `uvm_callback_iter` is also available, and is the generally
        preferred way to iterate over callback queues.
        Args:
            cls:
            obj:
            CB:
        Returns:
        """
        q = None
        if not cls.m_base_inst.m_pool.exists(obj):  # no instance specific
            if obj is None:
                q = cls.m_t_inst.m_tw_cb_q
            else:
                q = cls.m_t_inst.m_get_tw_cb_q(obj, CB)
        else:
            q = cls.m_base_inst.m_pool.get(obj)
            if q is None:
                q = UVMQueue()
                cls.m_base_inst.m_pool.add(obj, q)
        return q

    @classmethod
    def get_first(cls, itr, obj, CB=None):
        """
         Function: get_first

         Returns the first enabled callback of type CB which resides in the queue for `obj`.
         If `obj` is `None` then the typewide queue for T is searched. `itr` is the iterator
         it will be updated with a value that can be supplied to `get_next` to get the next
         callback object.

         If the queue is empty then `None` is returned.

         The iterator class `uvm_callback_iter` may be used as an alternative, simplified,
         iterator interface.

        Args:
            cls:
            itr:
            obj:
            CB:
        Returns:
        """
        cls.get()
        # tpoikela: added to mimic typed cbs
        typed_cls = cls._get_typed_cbs(obj, CB)
        cls = typed_cls
        q = cls.m_get_q(obj, CB)
        for i in range(len(q)):
            itr.m_i = i
            cb = []
            if CB is not None:
                if sv.cast(cb, q[itr.m_i], CB) and cb[0].callback_mode():
                    return cb[0]
            else:
                return q[itr.m_i]  # Just return 1st index without type match
        return None
        #  endfunction


    @classmethod
    def get_last(cls, itr, obj, CB=None):
        """
         Function: get_last

         Returns the last enabled callback of type CB which resides in the queue for `obj`.
         If `obj` is `None` then the typewide queue for T is searched. `itr` is the iterator
         it will be updated with a value that can be supplied to `get_prev` to get the previous
         callback object.

         If the queue is empty then `None` is returned.

         The iterator class `uvm_callback_iter` may be used as an alternative, simplified,
         iterator interface.

         static function CB get_last (ref int itr, input T obj)
        Args:
            cls:
            itr:
            obj:
            CB:
        Returns:
        """
        cls.get()
        # tpoikela: added to mimic typed cbs
        typed_cls = cls._get_typed_cbs(obj, CB)
        cls = typed_cls
        q = cls.m_get_q(obj, CB)
        for i in range(len(q)-1, -1, -1):
            itr.m_i = i
            cb = []
            if CB is not None:
                if sv.cast(cb, q[itr.m_i], CB) and cb[0].callback_mode():
                    return cb[0]
            else:
                return q[itr.m_i]  # Just return 1st index without type match
        return None

    @classmethod
    def get_next(cls, itr, obj, CB=None):
        """
         Function: get_next

         Returns the next enabled callback of type CB which resides in the queue for `obj`,
         using `itr` as the starting point. If `obj` is `None` then the typewide queue for T
         is searched. `itr` is the iterator; it will be updated with a value that can be
         supplied to `get_next` to get the next callback object.

         If no more callbacks exist in the queue, then `None` is returned. `get_next` will
         continue to return `None` in this case until `get_first` or `get_last` has been used to reset
         the iterator.

         The iterator class `uvm_callback_iter` may be used as an alternative, simplified,
         iterator interface.

         static function CB get_next (ref int itr, input T obj)
        Args:
            cls:
            itr:
            obj:
            CB:
        Returns:
        """
        cls.get()
        # tpoikela: added to mimic typed cbs
        typed_cls = cls._get_typed_cbs(obj, CB)
        cls = typed_cls
        q = cls.m_get_q(obj)
        for i in range(itr.m_i + 1, len(q)):
            itr.m_i = i
            cb = []
            if CB is not None:
                if sv.cast(cb, q[itr.m_i], CB) and cb[0].callback_mode():
                    return cb[0]
            else:
                return q[itr.m_i]
        return None
        #  endfunction

    @classmethod
    def get_prev(cls, itr, obj, CB=None):
        """


         Function: get_prev

         Returns the previous enabled callback of type CB which resides in the queue for `obj`,
         using `itr` as the starting point. If `obj` is `None` then the typewide queue for T
         is searched. `itr` is the iterator; it will be updated with a value that can be
         supplied to `get_prev` to get the previous callback object.

         If no more callbacks exist in the queue, then `None` is returned. `get_prev` will
         continue to return `None` in this case until `get_first` or `get_last` has been used to reset
         the iterator.

         The iterator class `uvm_callback_iter` may be used as an alternative, simplified,
         iterator interface.

         static function CB get_prev (ref int itr, input T obj)
        Args:
            cls:
            itr:
            obj:
            CB:
        Returns:
        """
        cls.get()
        # tpoikela: added to mimic typed cbs
        typed_cls = cls._get_typed_cbs(obj, CB)
        cls = typed_cls
        q = cls.m_get_q(obj)
        # for(itr = itr-1; itr>= 0; --itr)
        for i in range(itr.m_i - 1, -1, -1):
            itr.m_i = i
            cb = []
            if CB is not None:
                if sv.cast(cb, q[itr.m_i], CB) and cb[0].callback_mode():
                    return cb[0]
            else:
                return q[itr.m_i]
        return None


    @classmethod
    def display(cls, obj=None):
        """
         #-------------
         Group: Debug
         #-------------

         Function: display

         This function displays callback information for `obj`. If `obj` is
         `None`, then it displays callback information for all objects
         of type `T`, including typewide callbacks.

         static function void display(T obj=None)
        Args:
            cls:
            obj:
        """
        # For documentation purposes, need a function wrapper here.
        cls.get()
        super().display(obj)

    #endclass


#------------------------------------------------------------------------------
#
# Class- uvm_derived_callbacks #(T,ST,CB)
#
#------------------------------------------------------------------------------
# This type is not really expected to be used directly by the user, instead they are
# expected to use the macro `uvm_set_super_type. The sole purpose of this type is to
# allow for setting up of the derived_type/super_type mapping.
#------------------------------------------------------------------------------

#class uvm_derived_callbacks#(type T=uvm_object, type ST=uvm_object, type CB=uvm_callback)
#    extends uvm_callbacks#(T,CB)
#
#  typedef uvm_derived_callbacks#(T,ST,CB) this_type
#  typedef uvm_callbacks#(T)            this_user_type
#  typedef uvm_callbacks#(ST)           this_super_type
#
#  # Singleton instance is used for type checking
#  static this_type m_d_inst
#  static this_user_type m_user_inst
#  static this_super_type m_super_inst
#
#  # typeinfo
#  static UVMTypeIDBase m_s_typeid
#
#  static function this_type get()
#    m_user_inst = this_user_type::get()
#    m_super_inst = this_super_type::get()
#    m_s_typeid = uvm_typeid#(ST)::get()
#    if(m_d_inst is None):
#      m_d_inst = new
#    end
#    return m_d_inst
#  endfunction
#
#  static function bit register_super_type(string tname="", sname="")
#    this_user_type u_inst = this_user_type::get()
#    this_type      inst = this_type::get()
#    uvm_callbacks_base s_obj
#
#    this_user_type::m_t_inst.m_typename = tname
#
#    if(sname != "") m_s_typeid.typename = sname
#
#    if(u_inst.m_super_type is not None):
#      if(u_inst.m_super_type == m_s_typeid) return 1
#      uvm_report_warning("CBTPREG", { "Type ", tname, " is already registered to super type ",
#        this_super_type::m_t_inst.m_typename, ". Ignoring attempt to register to super type ",
#        sname}, UVM_NONE)
#      return 1
#    end
#    if(this_super_type::m_t_inst.m_typename == "")
#      this_super_type::m_t_inst.m_typename = sname
#    u_inst.m_super_type = m_s_typeid
#    u_inst.m_base_inst.m_super_type = m_s_typeid
#    s_obj = UVMTypeIDBase::typeid_map[m_s_typeid]
#    s_obj.m_derived_types.push_back(m_typeid)
#    return 1
#  endfunction
#
#endclass

class UVMCallbackIter:  # (type T = uvm_object, type CB = uvm_callback)
    """
    The ~uvm_callback_iter~ class is an iterator class for iterating over
    callback queues of a specific callback type. The typical usage of
    the class is:

    .. code-block:: python

      uvm_callback_iter#(mycomp,mycb) iter = new(this)
      for(mycb cb = iter.first(); cb is not None; cb = iter.next())
        cb.dosomething()

    The callback iteration macros, <`uvm_do_callbacks> and
    <`uvm_do_callbacks_exit_on> provide a simple method for iterating
    callbacks and executing the callback methods.
    """

    def __init__(self, obj, CB=None):
        """
        Function: new

        Creates a new callback iterator object. It is required that the object
        context be provided.
        Args:
            obj:
            CB:
        """
        self.m_i = 0
        self.m_cb: Optional[UVMCallback] = None  # UVMCallback
        self.m_obj = obj
        self.CB = CB

    def first(self) -> Optional['UVMCallback']:
        """
        Function: first

        Returns the first valid (enabled) callback of the callback type (or
        a derivative) that is in the queue of the context object. If the
        queue is empty then `None` is returned.
        Returns:
        """
        self.m_cb = UVMCallbacks.get_first(self, self.m_obj, self.CB)
        return self.m_cb

    def last(self) -> Optional['UVMCallback']:
        """
        Function: last

        Returns the last valid (enabled) callback of the callback type (or
        a derivative) that is in the queue of the context object. If the
        queue is empty then `None` is returned.
        Returns:
        """
        self.m_cb = UVMCallbacks.get_last(self, self.m_obj, self.CB)
        return self.m_cb

    def next(self) -> Optional['UVMCallback']:
        """
        Returns the next valid (enabled) callback of the callback type (or
        a derivative) that is in the queue of the context object. If there
        are no more valid callbacks in the queue, then `None` is returned.

        Returns:
        """
        self.m_cb = UVMCallbacks.get_next(self, self.m_obj, self.CB)
        return self.m_cb

    def prev(self) -> Optional['UVMCallback']:
        """
        Returns the previous valid (enabled) callback of the callback type (or
        a derivative) that is in the queue of the context object. If there
        are no more valid callbacks in the queue, then `None` is returned.

        Returns:
        """
        self.m_cb = UVMCallbacks.get_prev(self, self.m_obj, self.CB)
        return self.m_cb

    def get_cb(self) -> Optional['UVMCallback']:
        """
        Returns the last callback accessed via a first() or next()
        call.

        Returns:
        """
        return self.m_cb

    #function void trace(uvm_object obj = None)
    #   if (m_cb is not None && T::cbs::get_debug_flags() & UVM_CALLBACK_TRACE):
    #      uvm_report_object reporter = None
    #      string who = "Executing "
    #      void'($cast(reporter, obj))
    #      if (reporter is None) void'($cast(reporter, m_obj))
    #      if (reporter is None) reporter = uvm_top
    #      if (obj is not None) who = {obj.get_full_name(), " is executing "}
    #      elif (m_obj is not None) who = {m_obj.get_full_name(), " is executing "}
    #      reporter.uvm_report_info("CLLBK_TRC", {who, "callback ", m_cb.get_name()}, UVM_LOW)
    #   end
    #endfunction
    #endclass


class UVMCallback(UVMObject):
    """
    The `UVMCallback` class is the base class for user-defined callback classes.
    Typically, the component developer defines an application-specific callback
    class that extends from this class. In it, he defines one or more virtual
    methods, called a ~callback interface~, that represent the hooks available
    for user override.

    Methods intended for optional override should not be declared ~pure.~ Usually,
    all the callback methods are defined with empty implementations so users have
    the option of overriding any or all of them.

    The prototypes for each hook method are completely application specific with
    no restrictions.
    """


    reporter = UVMReportObject("cb_tracer")

    def __init__(self, name="uvm_callback"):
        """
        Creates a new uvm_callback object, giving it an optional `name`.

        Args:
            name (str): Name of the callback
        """
        super().__init__(name)
        self.m_enabled = True

    def callback_mode(self, on=-1):
        """
        Enable/disable callbacks (modeled like rand_mode and constraint_mode).

        Args:
            on:
        Returns:
        """
        if on == 0 or on == 1:
            is_en = "DISABLED"
            if on == 1:
                is_en = "ENABLED"
            uvm_cb_trace_noobj(self,sv.sformatf("Setting callback mode for %s to %s",
                  self.get_name(), is_en))
        else:
            is_en = "DISABLED"
            if self.m_enabled is True:
                is_en = "ENABLED"
            uvm_cb_trace_noobj(self, sv.sformatf("Callback mode for %s is %s",
                self.get_name(), is_en))
        callback_mode = self.m_enabled
        if on == 0:
            self.m_enabled = False
        if on == 1:
            self.m_enabled = True
        return callback_mode

    def is_enabled(self):
        """
        Returns 1 if the callback is enabled, 0 otherwise.

        Returns:
            bool: 1 if the callback is enabled, 0 otherwise.
        """
        return self.callback_mode()

    type_name = "uvm_callback"

    def get_type_name(self):
        """
        Returns the type name of this callback object.

        Returns:
            str: Type name of this callback object.
        """
        return UVMCallback.type_name



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from uvm.base.sv import sv
from uvm.base.uvm_callback import *
from uvm.comps.uvm_monitor import UVMMonitor
from uvm.macros import *
from uvm.tlm1 import UVMAnalysisPort
from cocotb.triggers import *
from apb import *


class apb_monitor_cbs(UVMCallback):
    def trans_observed(self, xactor, cycle):
        pass


class apb_monitor(UVMMonitor):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.ap = UVMAnalysisPort("ap", self)  # uvm_analysis_port#(apb_rw)
        self.cfg = None  # apb_config
        self.tr = None  # apb_rw
        self.sigs = None
    #   endfunction: new

    def build_phase(self, phase):
        agent = []
        if (sv.cast(agent, self.get_parent(), apb_agent) and len(agent) == 1):
            self.sigs = agent[0].vif
        else:
            tmp = []
            if (not UVMConfigDb.get(self, "", "vif", tmp)):
                uvm_fatal("APB/MON/NOVIF",
                   "No virtual interface specified for self monitor instance")
            self.sigs = tmp[0]

    
    async def run_phase(self, phase):
        super().run_phase(phase)
        while True:
            tr = None
    
            # Wait for a SETUP cycle
            while True:
                await Edge(self.sigs.pck)
                if (self.sigs.pck.psel != 1 or
                   self.sigs.pck.penable != 0):
                    break
    
            tr = apb_rw.type_id.create("tr", self)
    
            tr.kind = apb_rw.READ
            if (self.sigs.pck.pwrite):
                tr.kind = apb_rw.WRITE
            tr.addr = self.sigs.pck.paddr
    
            await Edge(self.sigs.pck)
            if (self.sigs.pck.penable != 1):
                uvm_error("APB", "APB protocol violation: SETUP cycle not followed by ENABLE cycle")
            tr.data = self.sigs.pck.pwdata
            if (tr.kind == apb_rw.READ):
                tr.data = self.sigs.pck.prdata
    
            self.trans_observed(tr)
            uvm_do_callbacks(apb_monitor,apb_monitor_cbs,
                    self.trans_observed(self, tr))
            self.ap.write(tr)

    def trans_observed(self, tr):
        pass

    #endclass: apb_monitor
uvm_component_utils(apb_monitor)



""" Common code used in unit testing """

from uvm.reg.uvm_reg_block import UVMRegBlock
from uvm.reg.uvm_reg import UVMReg
from uvm.reg.uvm_reg_model import UVM_READ
from uvm.reg.uvm_reg_field import UVMRegField
from uvm.reg.uvm_reg_map import UVMRegMap


def create_reg(name, nbits=32, num_fields=1):
    reg = UVMReg(name, nbits)
    lsb_pos = 0
    nbits_per_fields = int(nbits / num_fields)
    msb_pos = nbits_per_fields - 1
    for i in range(num_fields):
        field = UVMRegField.type_id.create(name, None, reg.get_full_name())
        field.configure(reg, msb_pos, lsb_pos, "RW", 0, 0x00, 1, 1, 1)
        lsb_pos += nbits_per_fields
        msb_pos += nbits_per_fields
    return reg


def create_reg_block(name, num_regs=1):
    rb = UVMRegBlock(name)
    rb.default_map = rb.create_map("", 0, 4, 0, 0)
    offset = 0
    for i in range(num_regs):
        reg = create_reg("reg_" + str(i))
        reg.configure(rb, None)
        if rb.default_map is not None:
            rb.default_map.add_reg(reg, offset, "RW")
        offset += 4
    rb.configure(hdl_path='path.to.top')
    return rb

def create_reg_map(name):
    rb = create_reg_block('test_reg_block_for_map', 0)
    # rb.default_map = rb.create_map("", 0, 4, 0, 0)
    reg_map = rb.default_map
    return [reg_map, rb]

class TestPacket():

    def __init__(self, data=0, addr=0):
        self.data = data
        self.addr = addr

    def convert2string(self):
        return "Addr: " + str(self.addr)


class TestRegAdapter():

    def bus2reg(self, pkt, rw):
        rw.data = pkt.data
        rw.addr = pkt.addr
        rw.kind = UVM_READ
        return rw

    def reg2bus(self):
        pass


class MockObj:
    inst_id = 0

    """ Mock object for unit testing to reduce import deps between files """
    def __init__(self, name, parent):
        self.m_parent = parent
        self.m_name = name
        self.m_typename = "MockObj"
        self.m_inst_id = MockObj.inst_id
        MockObj.inst_id += 1

    def get_name(self):
        """Get name of this object """
        return self.m_name

    def get_full_name(self):
        """Get full name of this object """
        if self.m_parent is not None:
            return self.m_parent.get_full_name() + "." + self.m_name
        return self.m_name

    def get_parent(self):
        """Get parent of this object """
        return self.m_parent

    def get_type_name(self):
        return self.m_typename

    def get_inst_id(self):
        return self.m_inst_id


class MockCb(MockObj):

    def __init__(self, name):
        super().__init__(name, None)
        self.m_enabled = True

    def callback_mode(self, on=-1):
        return self.m_enabled




import unittest

from uvm.base.uvm_object import UVMObject
from uvm.macros.uvm_object_defines import *
from uvm.base.uvm_object_globals import *

class TestObj(UVMObject):

    def __init__(self, name):
        super().__init__(name)
        self.addr = 0
        self.data = 10

uvm_object_utils_begin(TestObj)
uvm_field_int("addr")
uvm_field_int("data", UVM_COMPARE)
uvm_object_utils_end(TestObj)



class SuperObj(UVMObject):

    def __init__(self, name):
        super().__init__(name)
        self.my_val = 123
        self.my_obj = TestObj("test_obj")
        self.my_obj.addr = 3
        self.my_obj.data = 123

uvm_object_utils_begin(SuperObj)
uvm_field_int("my_val")
uvm_field_object("my_obj")
uvm_object_utils_end(SuperObj)


class TestUVMObject(unittest.TestCase):

    def test_name(self):
        comp = UVMObject("my_component")
        self.assertEqual(comp.get_name(), "my_component")

    def test_inst_count(self):
        UVMObject.m_inst_count = 0
        comp = UVMObject("my_component")
        self.assertEqual(comp.get_inst_count(), 1)

    def test_object_utils(self):
        o1 = TestObj("o1")
        o1.addr = 10
        o2 = o1.clone()
        self.assertEqual(o2.addr, 10)
        o1.data = 20
        o3 = o1.clone()
        self.assertFalse(o1.compare(o3))
        self.assertTrue(o1.compare(o1))

    def test_hier_object_utils(self):
        sup_obj = SuperObj("super_obj")
        so2 = sup_obj.clone()
        so2.my_obj.data = 123
        self.assertTrue(sup_obj.compare(so2))
        so2.my_obj.data = 666
        self.assertFalse(sup_obj.compare(so2))

    def test_pack_unpack(self):
        o2 = TestObj("o1")
        o2.addr = 0x234
        o2.data = 567
        size, packed_obj = o2.pack()
        print("Packed obj is " + str(packed_obj) + ' size: ' + str(size))
        o3 = TestObj("o3")
        o3.unpack(packed_obj)
        self.assertEqual(o3.addr, 0x234)
        self.assertEqual(o3.data, 567)

        sup_obj = SuperObj("super_obj")
        sup_obj.my_obj.addr = 888
        size, packed_obj = sup_obj.pack()
        print(str(packed_obj))
        print("packed obj is " + hex(packed_obj))
        sup_obj22 = SuperObj("super_obj22")
        self.assertEqual(sup_obj22.my_obj.addr, 3)
        sup_obj22.unpack(packed_obj)
        self.assertEqual(sup_obj22.my_obj.addr, 888)


class TestRecordIntegration(unittest.TestCase):

    def test_record(self):
        from uvm.base.uvm_tr_database import UVMTextTrDatabase
        db = UVMTextTrDatabase()
        self.assertTrue(db.open_db())
        stream = db.open_stream("obj_str")
        self.assertTrue(stream.is_open())
        o2 = TestObj("o2")
        recorder = stream.open_recorder("obj_recorder")
        o2.record(recorder)
        sup_obj = SuperObj("super_obj")
        sup_obj.record(recorder)



if __name__ == '__main__':
    unittest.main()




import unittest
from uvm.base.uvm_report_server import UVMReportServer
from uvm.base.uvm_object_globals import *


class TestUVMReportServer(unittest.TestCase):

    def test_server(self):
        srv = UVMReportServer()

    def test_quit_count(self):
        srv = UVMReportServer()
        self.assertEqual(srv.get_quit_count(), 0)
        self.assertEqual(srv.is_quit_count_reached(), True)
        srv.set_max_quit_count(10, True)
        self.assertEqual(srv.is_quit_count_reached(), False)
        srv.set_quit_count(11)
        self.assertEqual(srv.is_quit_count_reached(), True)

    def test_severity_count(self):
        srv = UVMReportServer()
        self.assertEqual(srv.get_severity_count(UVM_ERROR), 0)
        srv.set_severity_count(UVM_FATAL, 10)
        self.assertEqual(srv.get_severity_count(UVM_FATAL), 10)
        srv.incr_severity_count(UVM_FATAL)
        self.assertEqual(srv.get_severity_count(UVM_FATAL), 11)
        srv.incr_severity_count(UVM_ERROR)
        self.assertEqual(srv.get_severity_count(UVM_ERROR), 1)

    def test_id_count(self):
        srv = UVMReportServer()
        self.assertEqual(srv.get_id_count("a"), 0)
        srv.incr_id_count("a")
        self.assertEqual(srv.get_id_count("a"), 1)
        srv.set_id_count("b", 10)
        self.assertEqual(srv.get_id_count("b"), 10)

    def test_do_print(self):
        srv = UVMReportServer()
        from uvm.base.uvm_printer import UVMTablePrinter
        printer = UVMTablePrinter()
        srv.do_print(printer)
        srv_as_str = printer.emit()
        self.assertRegex(srv_as_str, r'quit_count')
        self.assertRegex(srv_as_str, r'UVM_INFO')
        self.assertRegex(srv_as_str, r'UVM_ERROR')
        self.assertRegex(srv_as_str, r'UVM_FATAL')
        self.assertRegex(srv_as_str, r'UVM_WARNING')


    def test_process_report_message(self):
        srv = UVMReportServer()
        from uvm.base.uvm_report_message import UVMReportMessage
        rpt_msg = UVMReportMessage()
        rpt_msg.set_action(UVM_LOG | UVM_DISPLAY)
        srv.process_report_message(rpt_msg)

    def test_execute_report_message(self):
        srv = UVMReportServer()

    def test_compose_report_message(self):
        srv = UVMReportServer()

    def test_report_summarize(self):
        srv = UVMReportServer()
        from uvm.base.uvm_report_message import UVMReportMessage
        rpt_msg = UVMReportMessage()
        rpt_msg.set_action(UVM_LOG | UVM_DISPLAY)
        srv.process_report_message(rpt_msg)
        srv.report_summarize()

if __name__ == '__main__':
    unittest.main()



#
#------------------------------------------------------------------------------
# Copyright 2007-2011 Mentor Graphics Corporation
# Copyright 2017 Intel Corporation
# Copyright 2010 Synopsys, Inc.
# Copyright 2007-2018 Cadence Design Systems, Inc.
# Copyright 2010 AMD
# Copyright 2015-2018 NVIDIA Corporation
# Copyright 2020 Tuomas Poikela (tpoikela)
#
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------


from typing import Union, Generic, TypeVar, List, Optional, Any

from .uvm_object import UVMObject
from .uvm_globals import uvm_report_warning

T = TypeVar('T')


class UVMQueue(UVMObject, Generic[T]):
    """
    Implements a class-based dynamic queue.
    """

    type_name = "uvm_queue"
    m_global_queue: 'UVMQueue' = None

    def __init__(self, name=""):
        UVMObject.__init__(self, name)
        self.queue: List[T] = list()

    @classmethod
    def get_global_queue(cls) -> 'UVMQueue':
        """
        Function: get_global_queue

        Returns the singleton global queue.

        This allows items to be shared amongst components throughout the
        verification environment.

        Returns:
            UVMQueue: The global singleton queue
        """
        if UVMQueue.m_global_queue is None:
            UVMQueue.m_global_queue = UVMQueue("global_queue")
        return UVMQueue.m_global_queue

    @classmethod
    def get_global(cls, index) -> Any:
        """
        Function: get_global

        Returns the specified item instance from the global item queue.
        Args:
            cls:
            index:
        Returns:
        """
        gqueue = UVMQueue.get_global_queue()
        return gqueue.get(index)

    def get(self, index: int) -> T:
        """
        Function: get

        Returns the item at the given `index`.

        If no item exists by that key, a new item is created with that key
        and returned.
        Args:
            index:
        Returns:
        """
        default_value = 0
        if index >= self.size() or index < 0:
            uvm_report_warning("QUEUEGET",
                "get: given index out of range for queue of size {}. Ignoring get request"
                .format(self.size()))
            return default_value
        return self.queue[index]

    def size(self) -> int:
        """
        Function: size

        Returns the number of items stored in the queue.
        Returns:
        """
        return len(self.queue)

    def __len__(self) -> int:
        """
        len() operator
        Returns:
        """
        return self.size()

    def __setitem__(self, i: int, value):
        """
        Implements aa[x] = y
        Args:
            i:
            value:
        Raises:
        """
        if i < self.size():
            self.queue[i] = value
        else:
            raise Exception("UVMQueue set index {} ouf of bounds (size: {})".format(
                i, self.size()))


    def __getitem__(self, i: Union[int, slice]) -> Union[T, List[T]]:
        """
        Implements aa[x]
        Args:
            i:
        Returns:
        Raises:
        """
        if isinstance(i, slice):
            # TODO should handle slices like [:] and [:-1]
            res = []
            for i in range(slice.start, slice.stop):
                res.append(self.queue[i])
            return res
        elif i < self.size():
            return self.queue[i]
        else:
            raise IndexError("UVMQueue get index {} ouf of bounds (size: {})".format(
                i, self.size()))

    def insert(self, index: int, item: T) -> None:
        """
        Function: insert

        Inserts the item at the given `index` in the queue.
        Args:
            index:
            item:
        """
        if index >= self.size() or index < 0:
            uvm_report_warning("QUEUEINS",
                "insert: given index {} out of range for queue of size {}. Ignoring insert request"
                .format(index, self.size()))
            return
        self.queue.insert(index, item)

    def delete(self, index=-1) -> None:
        """
        Function: delete

        Removes the item at the given `index` from the queue; if `index` is
        not provided, the entire contents of the queue are deleted.
        Args:
            index:
        """
        if index >= self.size() or index < -1:
            uvm_report_warning("QUEUEDEL",
                "delete: given index out of range for queue of size {}. Ignoring delete request"
                .format(self.size()))
            return
        if index == -1:
            self.queue = list()
        else:
            self.queue.pop(index)

    def pop_front(self) -> T:
        """
        Function: pop_front

        Returns the first element in the queue (index=0),
        or `null` if the queue is empty.
        Returns:
        Raises:
        """
        if self.size() > 0:
            val = self.queue[0]
            del self.queue[0]
            return val
        else:
            raise Exception('pop_front() called on empty queue')

    def front(self) -> Optional[T]:
        if self.size() > 0:
            return self.queue[0]
        return None

    def back(self) -> Optional[T]:
        if self.size() > 0:
            return self.queue[len(self.queue) - 1]
        return None

    def pop_back(self) -> Optional[T]:
        """
        Function: pop_back

        Returns the last element in the queue (index=size()-1),
        or `null` if the queue is empty.
        Returns:
        """
        return self.queue.pop()

    def push_front(self, item: T) -> None:
        """
        Function: push_front

        Inserts the given `item` at the front of the queue.
        Args:
            item:
        """
        self.queue.insert(0, item)

    def push_back(self, item: T) -> None:
        """
        Function: push_back

        Inserts the given `item` at the back of the queue.
        Args:
            item:
        """
        self.queue.append(item)

    def create(self, name="") -> 'UVMQueue[T]':
        v = UVMQueue(name)
        return v

    def get_type_name(self) -> str:
        return UVMQueue.type_name

    def do_copy(self, rhs) -> None:
        if rhs is None:
            return
        UVMObject.do_copy(self, rhs)
        self.queue = rhs.queue

    def convert2string(self) -> str:
        return str(self.queue)

    def __str__(self) -> str:
        return self.convert2string()

    def find_with(self, find_func) -> 'UVMQueue[T]':
        """
        Group: find functions
        Added by tpoikela to mimic SystemVerilog find API
        Args:
            find_func:
        Returns:
        """
        qq = UVMQueue()
        for ee in self.queue:
            if find_func(ee):
                qq.push_back(ee)
        return qq

    def find_first_index(self, find_func) -> int:
        idx = -1
        for i in range(len(self.queue)):
            ee = self.queue[i]
            if find_func(ee):
                idx = i
                break
        return idx



#//----------------------------------------------------------------------
#//   Copyright 2010 Mentor Graphics Corporation
#//   Copyright 2010-2011 Synopsys, Inc
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------


import cocotb
from cocotb.triggers import Timer
from uvm import *
from apb_rw import apb_rw


class target(UVMComponent):

    #   local bit [31:0] m_data
    #   uvm_tlm_b_target_socket #(target, apb_rw) sock


    def __init__(self, name="target", parent=None):
        super().__init__(name, parent)
        self.sock = UVMTLMBTargetSocket("sock", self)
        self.m_data = 0xDEADBEEF
        self.num_transport = 0


    
    async def b_transport(self, rw, delay):
        if rw.addr == 0x0000FF00:
            delay.incr(5, 1)
            if rw.kind == apb_rw.READ:
                rw.data = self.m_data
            else:
                self.m_data = rw.data
        self.num_transport += 1
        await Timer(5, "NS")


    def start_of_simulation_phase(self, phase):
        uvm_info("TRGT/RPT/START", sv.sformatf("m_data: 'h%h", self.m_data), UVM_NONE)


    def report_phase(self, phase):
        uvm_info("TRGT/RPT/FINAL", sv.sformatf("m_data: 'h%h", self.m_data), UVM_NONE)


uvm_component_utils(target)



#//
#// -------------------------------------------------------------
#//    Copyright 2010-2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from cocotb_coverage.coverage import coverage_section, CoverPoint, CoverCross

from uvm.reg import (UVMReg, UVMRegField, UVMMem, UVMRegBlock,
    UVM_CVR_FIELD_VALS, UVM_CVR_REG_BITS, UVM_CVR_ADDR_MAP,
    UVM_BIG_ENDIAN)
from uvm.macros import uvm_object_utils

#//
#// This example demonstrates how to include a coverage model
#// in a register model.
#//
#// Three coverage models are included:
#//   - Register level coverage
#//   - Address-leel coverage
#//   - Field value coverage
#//
#// This example demonstrates *how* to integrate a coverage model
#// in a UVM Register model. The details of the coverage model itself
#// are generator-specific and outside the scope of UVM.
#//


def cond(is_read, byte_en):
    return is_read is False and ((byte_en & 0x1) == 1)


class reg_R(UVMReg):

    CgBits = coverage_section(
        CoverPoint('top.cb_bits.wF1_0',
            xf=lambda data, byte_en, is_read, m_curr: cond(is_read, byte_en) and (m_curr & data & (1 << 1)),
            bins=[True, False]
        ),  # {m_current[0],m_data[0]} iff (!m_is_read  and  m_be[0])
        #      @CoverPoint('wF1_1', xf = lambda a: a, bins = []) #  {m_current[1],m_data[1]} iff (!m_is_read  and  m_be[0])
        #      @CoverPoint('wF1_2', xf = lambda a: a, bins = []) #  {m_current[2],m_data[2]} iff (!m_is_read  and  m_be[0])
        #      @CoverPoint('wF1_3', xf = lambda a: a, bins = []) #  {m_current[3],m_data[3]} iff (!m_is_read  and  m_be[0])
        #      @CoverPoint('wF2_0', xf = lambda a: a, bins = []) #  {m_current[4],m_data[4]} iff (!m_is_read  and  m_be[0])
        #      @CoverPoint('wF2_1', xf = lambda a: a, bins = []) #  {m_current[5],m_data[5]} iff (!m_is_read  and  m_be[0])
        #      @CoverPoint('wF2_2', xf = lambda a: a, bins = []) #  {m_current[6],m_data[6]} iff (!m_is_read  and  m_be[0])
        #      @CoverPoint('wF2_3', xf = lambda a: a, bins = []) #  {m_current[7],m_data[7]} iff (!m_is_read  and  m_be[0])
    )  # Close coverage section


    #   cg_vals = coverage_section(
    #      @CoverPoint('F1', xf = lambda a: a, bins = []) #  F1.value[3:0]
    #      @CoverPoint('F2', xf = lambda a: a, bins = []) #  F2.value[3:0]
    #      @CoverCross('F1F2', items = [ F1, F2
    #   ) # Close coverage section


    def __init__(self, name="reg_R"):
        super().__init__(name, 8)
        self.has_cover = self.build_coverage(UVM_CVR_REG_BITS + UVM_CVR_FIELD_VALS)
        if self.has_coverage(UVM_CVR_REG_BITS):
            # self.cg_bits = covergroup()
            pass
        if self.has_coverage(UVM_CVR_FIELD_VALS):
            # self.cg_vals = covergroup()
            pass
        self.m_current = 0
        self.m_data = 0
        self.m_be = 0
        self.m_is_read = False

    def build(self):
        self.F1 = UVMRegField.type_id.create("F1",None, self.get_full_name())
        self.F1.configure(self, 4, 0, "RW", 0, 0x00, 1, 1, 1)
        self.rand('F1')
        self.F1.set_compare()

        self.F2 = UVMRegField.type_id.create("F2",None,self.get_full_name())
        self.F2.configure(self, 4, 4, "RO", 0, 0xA, 1, 1, 1)
        self.rand('F2')


    def sample(self, data, byte_en, is_read, _map):
        if self.get_coverage(UVM_CVR_REG_BITS):
            self.m_current = self.get()
            self.m_data    = data
            self.m_be      = byte_en
            self.m_is_read = is_read
            self.sample_bits_cg(data, byte_en, is_read, self.m_current)

    @CgBits
    def sample_bits_cg(self, data, byte_en, is_read, m_current):
        pass

    def sample_fields_cg(self):
        pass

    def sample_values(self):
        super().sample_values()
        if self.get_coverage(UVM_CVR_FIELD_VALS):
            self.sample_fields_cg()

uvm_object_utils(reg_R)



class mem_M(UVMMem):


    CgAddr = coverage_section(
        CoverPoint(
            'top.cg_addr.MIN_MID_MAX',
            # *Note that lambda args must match sample() args (omit self)
            xf=lambda offset, is_read, _map: offset,
            bins=[(0), (1,126), (127)],
            bins_labels=['MIN', 'MID', 'MAX']
        ),
        CoverPoint(
            'top.cg_addr.read_write',
            # *Note that lambda args must match sample() args (omit self)
            xf=lambda offset, is_read, _map: is_read,
            bins=[True, False],
            bins_labels=['READ', 'WRITE']
        )
    )


    def __init__(self, name="mem_M"):
        super().__init__(name, 128, 8, "RW")
        self.m_offset = 0
        self.has_cover = self.build_coverage(UVM_CVR_ADDR_MAP)
        if self.has_coverage(UVM_CVR_ADDR_MAP):
            pass
            #self.cg_addr = covergroup()


    @CgAddr
    # *Note that lambda args must match sample() args (omit self)
    def sample(self, offset, is_read, _map):
        if self.get_coverage(UVM_CVR_ADDR_MAP):
            self.m_offset  = offset


uvm_object_utils(mem_M)


class block_B(UVMRegBlock):

    CgAddrRange = coverage_section(
        CoverPoint(
            'top.reg_block_range.Ra',
            xf=lambda addr: addr,
            bins=[0x0000]
        ),
        CoverPoint('top.reg_block_range.Rb',
            xf=lambda addr: addr,
            bins=[0x0100]
        ),
        CoverPoint('top.reg_block_range.mem',
            xf=lambda addr: addr,
            bins=[0x2000, (0x2001, 0x23FE), 0x23FF],
            bins_labels=['MIN', 'MID', 'MAX']
        )
    )

    CgVals = coverage_section(
        CoverPoint('top.reg_block_vals.Ra',
            xf=lambda ra, rb: ra,  # Ra.F1.value[3:0]
            bins=range(16)),
        CoverPoint('top.reg_block_vals.Rb',
            xf=lambda ra, rb: rb,  # Rb.F1.value[3:0]
            bins=range(16)),
        CoverCross('top.reg_block_B.RaRb', items=['top.reg_block_vals.Ra',
            'top.reg_block_vals.Rb'])
    )  # Close coverage section


    def __init__(self, name="B"):
        super().__init__(name)
        self.has_cover = self.build_coverage(UVM_CVR_ADDR_MAP+UVM_CVR_FIELD_VALS)
        self.m_offset = 0
        if self.has_coverage(UVM_CVR_ADDR_MAP):
            pass
            # self.cg_addr = covergroup()
        if self.has_coverage(UVM_CVR_FIELD_VALS):
            pass
            #self.cg_vals = covergroup()


    def build(self):

        self.default_map = self.create_map("", 0, 4, UVM_BIG_ENDIAN, 0)

        self.Ra = reg_R.type_id.create("Ra",None,self.get_full_name())
        self.Ra.configure(self, None)
        self.Ra.build()
        self.rand('Ra')

        self.Rb = reg_R.type_id.create("Rb",None,self.get_full_name())
        self.Rb.configure(self, None)
        self.Rb.build()
        self.rand('Rb')

        self.M = mem_M.type_id.create("M",None,self.get_full_name())
        self.M.configure(self)

        self.default_map.add_reg(self.Ra, 0x0000, "RW")
        self.default_map.add_reg(self.Rb, 0x0100, "RW")
        self.default_map.add_mem(self.M, 0x2000, "RW")


    def sample(self, offset, is_read, _map):
        if self. get_coverage(UVM_CVR_ADDR_MAP):
            self.m_offset  = offset
            self.sample_addr_cg(offset)


    @CgAddrRange
    def sample_addr_cg(self, offset):
        pass

    def sample_vals_cg(self, ra, rb):
        pass

    def sample_values(self):
        super().sample_values()
        if self.get_coverage(UVM_CVR_FIELD_VALS):
            ra = self.Ra.F1.value
            rb = self.Rb.F1.value
            self.sample_vals_cg(ra, rb)


uvm_object_utils(block_B)



#
#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

#
#About: uvm_tlm_fifo
#
# This test case supposed to test a producer consumer connection through a uvm_tlm_fifo of 10 levels.
#
# Walk through the test:
# Inside a module, there will be two threads established, the first one the *producer*
# will put a packet of an integer to put port, the second one *consumer* will get self
# packet through a get port. The put and get ports will be connected to the uvm_tlm_fifo exports. 
#
# the method *fifo.used* will be used here to display which fifo level is used.

import cocotb
from cocotb.triggers import Timer

# You can also do 'from uvm import *' but this gives lint errors for undefined names
from uvm import (sv, UVMComponent, UVMTest, uvm_component_utils, run_test)
from uvm.tlm1 import *


class packet():
    def __init__(self, v):
        self.i = v


class producer(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name,parent)
        self.data_out = UVMPutPort("data_out", self)

    
    async def run_phase(self, phase):
        await Timer(1, "NS")
        p = packet(0)

        while self.data_out.try_put(p):
            sv.display("%0t: put data %0d", sv.time(), p.i)
            await Timer(10, "NS")
            p = packet(p.i + 1)

        sv.display("try_put status return: %0d", p.i)
        sv.display("%0t: do a blocking put", sv.time())
        await self.data_out.put(p)
        sv.display("%0t: blocking put succeeded", sv.time())



class consumer(UVMComponent):


    def __init__(self, name, parent):
        super().__init__(name,parent)
        self.data_in = UVMGetPort("data_in", self)


    
    async def run_phase(self, phase):
        #100;  // fifo will fill up
        await Timer(100, "NS")  # FIFO will fill up
        sv.display("%0t: getting one", sv.time())

        p = []
        await self.data_in.get(p)
        p = p[0]  # Required for python-tlm now
        sv.display("%0t: received data %0d", sv.time(), p.i)
        #100;  // let the blocking put succeed
        p = []
        while self.data_in.try_get(p):
            p = p[0]
            sv.display("%0t: received data %0d", sv.time(), p.i)
            #10
            await Timer(10, "NS")
            p = []


class test(UVMTest):

    def __init__(self, name="", parent=None):
        super().__init__(name, parent)

    async def run_phase(self, phase):
        phase.raise_objection(None)
        #5us
        await Timer(5, "US")
        phase.drop_objection(None)
uvm_component_utils(test)


async def print_proc(fifo):
    for i in range(30):
        sv.display("%0t:   FIFO level %0d of %0d", sv.time(), fifo.used(), fifo.size())
        await Timer(10, "NS")
        #10


@cocotb.test()
async def module_top(dut):
    prod = producer("prod", None)
    cons = consumer("cons", None)
    fifo = UVMTLMFIFO("fifo", None, 10)

    prod.data_out.connect(fifo.put_export)
    cons.data_in.connect(fifo.get_export)

    proc1 = cocotb.fork(run_test("test"))
    proc2 = cocotb.fork(print_proc(fifo))
    await sv.fork_join([proc1, proc2])

#  initial begin
#    prod.data_out.connect(fifo.put_export)
#    cons.data_in.connect(fifo.get_export)
#
#    fork
#      run_test("test")
#    join
#  end
#
#endmodule




import unittest
from uvm.base.uvm_report_catcher import UVMReportCatcher


class MyReportCatcher(UVMReportCatcher):

    def __init__(self, name='catcher'):
        super().__init__(name)

    def catch(self):
        self.sev = self.get_severity()
        self._id = self.get_id()


class TestUVMReportCatcher(unittest.TestCase):

    def test_catcher(self):
        rpt_catcher = MyReportCatcher('catcher')


if __name__ == '__main__':
    unittest.main()



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer

from uvm.comps.uvm_env import UVMEnv
from uvm import (UVMConfigDb, sv, uvm_top, uvm_info, uvm_error, UVM_MEDIUM)
from classA import ClassA
from classB import ClassB


class my_env(UVMEnv):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.debug = 0
        self.inst1 = None
        self.inst2 = None


    def build_phase(self, phase):
        super().build_phase(phase)
        arr = []
        if UVMConfigDb.get(self, "", "debug", arr) is True:
            self.debug = arr[0]
        UVMConfigDb.set(self, "inst1.u2", "v", 5)
        UVMConfigDb.set(self, "inst2.u1", "v", 3)
        UVMConfigDb.set(self, "inst1.*", "s", 0x10)

        sv.display("%s: In Build: debug = %0d", self.get_full_name(), self.debug)

        self.inst1 = ClassA("inst1", self)
        self.inst2 = ClassB("inst2", self)
        if UVMConfigDb.get(uvm_top, "topenv", "should_match", arr):
            uvm_info("MATCH_OK", "should_match found from DB", UVM_MEDIUM)
        else:
            uvm_error("MATCH_ERROR", "should_match NOT found from DB")


    def get_type_name(self):
        return "my_env"

    def do_print(self, printer):
        printer.print_field("debug", self.debug, 1)


    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        uvm_top.print_topology()
        await Timer(10, "NS")
        phase.drop_objection(self)



#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Synopsys, Inc
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base.uvm_component import UVMComponent
from uvm.macros import *

from host import host
from device import device


class tb_env(UVMComponent):


    def __init__(self, name="tb_env", parent=None):
        super().__init__(name, parent)
        self.hst = None
        self.dev = None


    def build_phase(self, phase):
        self.hst = host.type_id.create("hst", self)
        self.dev = device.type_id.create("dev", self)


    def connect_phase(self, phase):
        self.hst.sock.connect(self.dev.sock)




import unittest

from uvm.tlm1.uvm_analysis_port import (UVMAnalysisImp, UVMAnalysisExport,
    UVMAnalysisPort)


class TestUVMAnalysisPort(unittest.TestCase):

    def test_ports(self):
        from uvm.base.uvm_domain import UVMDomain
        from uvm.base.uvm_component import UVMComponent
        domains = UVMDomain.get_common_domain()

        class MyComp(UVMComponent):
            def build(self):
                self.analysis_imp = UVMAnalysisImp('ap', self)

            def write(self, t):
                self.written = True
                self.t = t

        imp = UVMComponent("port_parent", None)
        analysis_port = UVMAnalysisPort('aport', None)
        analysis_export = UVMAnalysisExport('my_export_in_test1', imp)
        targetComp = MyComp('my_comp', None)
        targetComp.build()
        analysis_port.connect(analysis_export)
        analysis_export.connect(targetComp.analysis_imp)
        #analysis_port.connect(targetComp.analysis_imp)
        analysis_port.resolve_bindings()

        analysis_port.write(12345)
        self.assertEqual(targetComp.written, True)
        self.assertEqual(targetComp.t, 12345)


if __name__ == '__main__':
    unittest.main()




from .uvm_mem_access_seq import *
from .uvm_mem_walk_seq import *
from .uvm_reg_access_seq import *
from .uvm_reg_bit_bash_seq import *
from .uvm_reg_hw_reset_seq import *
from .uvm_reg_mem_built_in_seq import *
from .uvm_reg_mem_hdl_paths_seq import *
from .uvm_reg_mem_shared_access_seq import *



#------------------------------------------------------------------------------
#   Copyright 2007-2010 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------

from .uvm_object import UVMObject
from .uvm_object_globals import (UVM_DISPLAY, UVM_HEX, UVM_INFO, UVM_LOG,
    UVM_RM_RECORD, UVM_UNSIGNED)

#------------------------------------------------------------------------------
#
# CLASS: uvm_report_message_element_base
#
# Base class for report message element. Defines common interface.
#
#------------------------------------------------------------------------------


class UVMReportMessageElementBase(UVMObject):

    def __init__(self, name=""):
        UVMObject.__init__(self, name)
        self._action = UVM_DISPLAY
        self._name = ""
        self._val = None

    def set_name(self, name):
        self._name = name

    def get_name(self):
        return self._name

    def set_action(self, action):
        self._action = action

    def get_action(self):
        return self._action

    def set_val(self, val):
        self._val = val

    def get_val(self):
        return self._val

    def pprint(self, printer):
        if self._action & (UVM_LOG | UVM_DISPLAY):
            self.do_print(printer)

    def record(self, recorder):
        if self._action & UVM_RM_RECORD:
            self.do_record(recorder)

    def copy(self, rhs):
        self.do_copy(rhs)

    def clone(self):
        return self.do_clone()

    def do_print(self, printer):
        pass

    def do_record(self, recorder):
        pass

    def do_copy(self, rhs):
        pass

    def do_clone(self):
        pass

#------------------------------------------------------------------------------
#
# CLASS: uvm_report_message_int_element
#
# Message element class for integral type
#
#------------------------------------------------------------------------------
class UVMReportMessageIntElement(UVMReportMessageElementBase):
    pass

class UVMReportMessageStringElement(UVMReportMessageElementBase):
    pass

class UVMReportMessageObjectElement(UVMReportMessageElementBase):
    pass

#------------------------------------------------------------------------------
#
# CLASS: uvm_report_message_element_container
#
# A container used by report message to contain the dynamically added elements,
# with APIs to add and delete the elements.
#
#------------------------------------------------------------------------------


class UVMReportMessageElementContainer(UVMObject):

    def __init__(self, name="element_container"):
        UVMObject.__init__(self, name)
        self.elements = []

    def size(self):
        return len(self.elements)

    def delete(self, index):
        elements = []
        for i in range(0, self.size()):
            if i != index:
                elements.append(self.elements[i])
        self.elements = elements

    def delete_elements(self):
        self.elements = []

    def get_elements(self):
        return self.elements

    def add(self, name, val, action=UVM_LOG | UVM_RM_RECORD):
        elem = UVMReportMessageElementBase()
        elem.set_name(name)
        elem.set_val(val)
        elem.set_action(action)
        self.elements.append(elem)

    def do_print(self, printer):
        UVMObject.do_print(self, printer)
        for i in range(0, self.size()):
            self.elements[i].pprint(printer)

    def do_record(self, recorder):
        UVMObject.do_record(self, recorder)
        for i in range(0, self.size()):
            self.elements[i].record(recorder)

    def do_copy(self, rhs):
        super().do_copy(rhs)
        #uvm_report_message_element_container urme_container
        #if(!$cast(urme_container, rhs) || (rhs==null))
        #    return
        urme_container = rhs
        if rhs is None:
            return

        self.delete_elements()
        for i in range(len(urme_container.elements)):
            self.elements.append(urme_container.elements[i].clone())


#------------------------------------------------------------------------------
#
# CLASS: uvm_report_message
#
# The uvm_report_message is the basic UVM object message class.  It provides
# the fields that are common to all messages.  It also has a message element
# container and provides the APIs necessary to add integral types, strings and
# uvm_objects to the container. The report message object can be initialized
# with the common fields, and passes through the whole reporting system (i.e.
# report object, report handler, report server, report catcher, etc) as an
# object. The additional elements can be added/deleted to/from the message
# object anywhere in the reporting system, and can be printed or recorded
# along with the common fields.
#
#------------------------------------------------------------------------------


class UVMReportMessage(UVMObject):

    def __init__(self, name=""):
        UVMObject.__init__(self, name)

        self._report_object = None
        self._report_handler = None
        self._report_server = None

        self._severity = UVM_INFO
        self._id = ""
        self._message = ""
        self._verbosity = 0
        self._filename = ""
        self._line = ""
        self._context_name = ""
        self._action = UVM_LOG | UVM_DISPLAY
        self._file = 0
        self._report_message_element_container = UVMReportMessageElementContainer()

    @classmethod
    def new_report_message(cls, name="uvm_report_message") -> 'UVMReportMessage':
        """
        Creates a new uvm_report_message object.
        This function is the same as new(), but keeps the random stability.

        Args:
            name (str): Name of the message
        Returns:
            UVMReportMessage: Created report message
        """
        p = None
        # rand_state = ""
        #p = process::self()

        if p is not None:
            pass
            #rand_state = p.get_randstate()
        new_report_message = UVMReportMessage(name)
        # if (p != null):
        #   p.set_randstate(rand_state)
        return new_report_message

    def do_print(self, printer):
        """
        The uvm_report_message implements `UVMObject.do_print()` such that
        `print` method provides UVM printer formatted output
        of the message.  A snippet of example output is shown here:

        Args:
            printer (UVMPrinter): Printer used for printing the msg
        """
        l_verbosity = 0
        UVMObject.do_print(self, printer)
        printer.print_generic("severity", "uvm_severity",
                          32, str(self._severity))
        printer.print_string("id", self._id)
        printer.print_string("message", self._message)
        printer.print_int("verbosity", self._verbosity, 32, UVM_HEX)
        printer.print_string("filename", self._filename)
        printer.print_int("line", self._line, 32, UVM_UNSIGNED)
        printer.print_string("context_name", self._context_name)
        if self._report_message_element_container.size() != 0:
            self._report_message_element_container.print(printer)

    def do_copy(self, rhs):
        report_message = rhs
        UVMObject.do_copy(self, rhs)
        if rhs is None:
            return
        self._report_object = report_message.get_report_object()
        self._report_handler = report_message.get_report_handler()
        self._report_server = report_message.get_report_server()
        self._context_name = report_message.get_context()
        self._file = report_message.get_file()
        self._filename = report_message.get_filename()
        self._line = report_message.get_line()
        self._action = report_message.get_action()
        self._severity = report_message.get_severity()
        self._id = report_message.get_id()
        self._message = report_message.get_message()
        self._verbosity = report_message.get_verbosity()
        self._report_message_element_container.copy(report_message._report_message_element_container)

    def create(self, name="") -> 'UVMReportMessage':
        return UVMReportMessage(name)

    #----------------------------------------------------------------------------
    # Group:  Infrastructure References
    #----------------------------------------------------------------------------

    def get_report_object(self):
        """
        Function: get_report_object

        Returns:
            UVMReportObject: Report object for this msg.
        """
        return self._report_object

    def set_report_object(self, ro):
        """
        Get or set the uvm_report_object that originated the message.

        Args:
            ro (UVMReportObject):
        """
        self._report_object = ro

    def get_report_handler(self):
        """
        Get the UVMReportHandler that is responsible for checking
        whether the message is enabled, should be upgraded/downgraded, etc.

        Returns:
            UVMReportHandler: Report handler for this msg.
        """
        return self._report_handler

    def set_report_handler(self, rh):
        """
        Set the UVMReportHandler that is responsible for checking
        whether the message is enabled, should be upgraded/downgraded, etc.

        Args:
            rh:
        """
        self._report_handler = rh

    def get_report_server(self):
        """
        Get the `UVMReportServer` that is responsible for servicing
        the message's actions.

        Returns:
            UVMReportServer: Report server for this object
        """
        return self._report_server

    def set_report_server(self, rs):
        """
        Set the `UVMReportServer` that is responsible for servicing
        the message's actions.

        Args:
            rs (UVMReportServer): Report server for this object.
        """
        self._report_server = rs

    #----------------------------------------------------------------------------
    # Group:  Message Fields
    #----------------------------------------------------------------------------

    def get_severity(self):
        """
        Function: get_severity
        Returns:
        """
        return self._severity

    def set_severity(self, sev):
        """
        Function: set_severity

        Get or set the severity (UVM_INFO, UVM_WARNING, UVM_ERROR or
        UVM_FATAL) of the message.  The value of this field is determined via
        the API used (`uvm_info(), `uvm_waring(), etc.) and populated for the user.
        Args:
            sev:
        """
        self._severity = sev

    def get_id(self):
        """
        Function: get_id
        Returns:
        """
        return self._id

    def set_id(self, id):
        """
        Function: set_id

        Get or set the id of the message.  The value of this field is
        completely under user discretion.  Users are recommended to follow a
        consistent convention.  Settings in the uvm_report_handler allow various
        messaging controls based on this field.  See `uvm_report_handler`.
        Args:
            id:
        """
        self._id = id

    def get_message(self):
        """
        Function: get_message
        Returns:
        """
        return self._message

    def set_message(self, msg):
        """
        Function: set_message

        Get or set the user message content string.
        Args:
            msg:
        """
        self._message = msg

    def get_verbosity(self):
        """
        Function: get_verbosity
        Returns:
        """
        return self._verbosity

    def set_verbosity(self, ver):
        """
        Function: set_verbosity

        Get or set the message threshold value.  This value is compared
        against settings in the `uvm_report_handler` to determine whether this
        message should be executed.
        Args:
            ver:
        """
        self._verbosity = ver

    def get_filename(self):
        """
        Function: get_filename
        Returns:
        """
        return self._filename

    def set_filename(self, fname):
        """
        Function: set_filename

        Get or set the file from which the message originates.  This value
        is automatically populated by the messaging macros.
        Args:
            fname:
        """
        self._filename = fname

    def get_line(self):
        """
        Function: get_line
        Returns:
        """
        return self._line

    def set_line(self, ln):
        """
        Function: set_line

        Get or set the line in the `file` from which the message originates.
        This value is automatically populate by the messaging macros.
        Args:
            ln:
        """
        self._line = ln

    def get_context(self):
        """
        Get the optional user-supplied string that is meant to convey
        the context of the message.  It can be useful in scopes that are not
        inherently UVM like modules, interfaces, etc.

        Returns:
            str: Context for the msg.
        """
        return self._context_name

    def set_context(self, cn):
        """
        Set the optional user-supplied string that is meant to convey
        the context of the message.  It can be useful in scopes that are not
        inherently UVM like modules, interfaces, etc.

        Args:
            cn (str): Context for the msg.
        """
        self._context_name = cn

    def get_action(self):
        """
        Get the action(s) that the `UVMReportServer` should perform
        for this message.  This field is populated by the `UVMReportHandler` during
        message execution flow.

        Returns:
        """
        return self._action

    def set_action(self, act):
        """
        Set the action(s) that the `UVMReportServer` should perform
        for this message.  This field is populated by the `UVMReportHandler` during
        message execution flow.

        Args:
            act:
        """
        self._action = act

    def get_file(self):
        """
        Get or set the file that the message is to be written to when the
        message's action is UVM_LOG.  This field is populated by the
        `UVMReportHandler` during message execution flow.

        Returns:
            str: Filename associated with this msg.
        """
        return self._file

    def set_file(self, fl):
        """
        Set the file that the message is to be written to when the
        message's action is UVM_LOG.  This field is populated by the
        `UVMReportHandler` during message execution flow.

        Args:
            fl (str): Filename associated with this msg.
        """
        self._file = fl

    def get_element_container(self):
        """
        Get the element_container of the message

        Returns:
            UVMReportMessageElementContainer: Element_container of the message
        """
        return self._report_message_element_container

    def set_report_message(self, severity, id, message, verbosity, filename,
            line, context_name):
        """
        Set all the common fields of the report message in one shot.

        Args:
            severity:
            id:
            message:
            verbosity:
            filename:
            line:
            context_name:
        """
        self._context_name = context_name
        self._filename = filename
        self._line = line
        self._severity = severity
        self._id = id
        self._message = message
        self._verbosity = verbosity

    #----------------------------------------------------------------------------
    # Group-  Message Recording
    #----------------------------------------------------------------------------
    # TODO

    def add(self, name, value, action=UVM_LOG | UVM_RM_RECORD):
        """
        #----------------------------------------------------------------------------
        Group:  Message Element APIs
        #----------------------------------------------------------------------------
        Args:
            name:
            value:
            action:
        """
        self._report_message_element_container.add(name, value, action)




import unittest

from uvm.base.uvm_domain import UVMDomain
from uvm.base.uvm_common_phases import UVMBuildPhase
from uvm.base.uvm_phase import UVMPhase


class TestUVMDomain(unittest.TestCase):


    def test_common(self):
        common = UVMDomain.get_common_domain()
        self.assertNotEqual(common, None)
        bld = common.find(UVMBuildPhase.get())
        self.assertNotEqual(bld, None)


    def test_add_and_find(self):
        my_ph = UVMPhase('my_ph')
        dm = UVMDomain('my domain')
        dm.add(my_ph)
        my_ph2 = dm.find(my_ph)
        self.assertNotEqual(my_ph2, None)
        self.assertEqual(my_ph2.get_name(), 'my_ph')


if __name__ == '__main__':
    unittest.main()



#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//


#class vip_monitor_cbs(uvm_callback):
    #    def rxed(self,vip_monitor xactor, ref bit [7:0] chr):;endfunction


from cocotb.triggers import RisingEdge, FallingEdge

from uvm import *


class vip_monitor(UVMMonitor):


    #   `uvm_register_cb(vip_monitor, vip_monitor_cbs)
    #
    #   vip_rx_vif vif
    #
    #   local process m_proc
    #
    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.ap = UVMAnalysisPort("ap", self)

        self.m_in_sync   = 0
        self.m_suspend   = 1
        self.m_suspended = 1
        self.m_proc = None
        self.m_evt_sync_changed = Event("evt_sync_changed")
        self.hier_objection = False

    def is_in_sync(self):
        return self.m_in_sync

    async def wait_for_sync_change(self):
        await self.m_evt_sync_changed.wait()
        self.m_evt_sync_changed.clear()


    def build_phase(self, phase):
        vif = []
        if not UVMConfigDb.get(self, "", "vif", vif):
            uvm_fatal("VIP/MON/NOVIF", "No virtual interface specified for self monitor instance")
        self.vif = vif[0]

    async def reset_phase(self, phase):
        if self.hier_objection:
            phase.raise_objection(self, "Resetting monitor")
        uvm_info("VIP_MON", "reset_phase started", UVM_MEDIUM)
        await self.reset_and_suspend()
        if self.hier_objection:
            phase.drop_objection(self, "VIP monitor has been reset")

    async def reset_and_suspend(self):
        """
        Abruptly interrupt and suspend this monitor
        """
        self.m_suspend = 1
        if self.m_proc is not None:
            self.m_proc.kill()
            self.m_proc = None
            #wait (m_suspended)
            while self.m_suspended != 1:
                await RisingEdge(self.vif.clk)

    async def suspend(self):
        await self.reset_and_suspend()


    async def resume(self):
        self.m_suspend = 0
        while self.m_suspended != 0:
            await RisingEdge(self.vif.clk)


    def set_in_sync(self, val):
        self.m_in_sync = val
        self.m_evt_sync_changed.set()

    async def run_phase(self, phase):
        while True:
            self.m_suspended = 1
            self.set_in_sync(0)

            #wait (!m_suspend)
            while self.m_suspend != 0:
                await RisingEdge(self.vif.clk)
            self.m_suspended = 0
            self.m_proc = cocotb.fork(self.run_phase_fork(phase))
            await sv.fork_join([self.m_proc])


    async def run_phase_fork(self, phase):
        symbol = 0
        #m_proc = process::self();
        #
        while True:
            ok = 0
            escaped = 0

            uvm_info("VIP/MON/SYM", "Looking for SYNC", UVM_MEDIUM)

            # First, acquire symbol sync
            symbol = 0x00
            while symbol != 0xB2:
                await FallingEdge(self.vif.clk)
                #symbol = {symbol[6:0], vif.Rx}
                #symbol = (symbol << 1) | (0x1 & self.vif.Rx.value)
                symbol = self.get_symbol(symbol)

            uvm_info("VIP/MON/SYM1", "Found first SYNC", UVM_MEDIUM)

            # Must now find 3 more, 7 symbols apart
            ok = 1
            num_sync = 0
            for _ in range(3):
                for _ in range(7 * 8):
                    await FallingEdge(self.vif.clk)
                    #symbol = {symbol[6:0], vif.Rx}
                    symbol = self.get_symbol(symbol)

                if symbol == 0xB2:
                    num_sync += 1
                    uvm_info("VIP/MON/SYM/FR", "Found another SYNC, " + str(num_sync), UVM_MEDIUM)
                else:
                    uvm_info("VIP/MON/SYM/FR", "SYNC not found. Got: " + hex(symbol), UVM_MEDIUM)
                    ok = 0
                    break
            if ok:
                self.set_in_sync(1)
            else:
                uvm_info("VIP/MON/SYNC/NOT_ACQ", "SYNC NOT acquired!", UVM_MEDIUM)
                break

            uvm_info("VIP/MON/SYNC/ACQ", "SYNC acquired!", UVM_MEDIUM)
            escaped = 0

            while self.m_in_sync:
                for _ in range(6):
                    for _ in range(8):
                        await FallingEdge(self.vif.clk)
                        #symbol = {symbol[6:0], vif.Rx}
                        symbol = self.get_symbol(symbol)
                    if escaped or (symbol != 0x81 and symbol != 0xE7):
                        # vip_tr tr

                        self.rxed(symbol)
                        # TODO uvm_do_callbacks(vip_monitor, vip_monitor_cbs, rxed(self, symbol))
                        uvm_info("VIP/MON/RX", sv.sformatf("Received 0x%h", symbol),
                            UVM_HIGH)

                        tr = vip_tr.type_id.create("tr",None,self.get_full_name())
                        tr.chr = symbol
                        self.ap.write(tr)

                        escaped = 0
                    elif symbol == 0xE7:
                        escaped = 1
                    else:
                        print("vip_mon got symbol: {:x}".format(symbol))

                # Check that we are still in SYNC
                for _ in range(8):
                    await FallingEdge(self.vif.clk)
                    symbol = self.get_symbol(symbol)
                if symbol != 0xB2:
                    self.set_in_sync(0)
                    uvm_warning("VIP/MON/SYNC/LOST", "SYNC lost!")

    def get_symbol(self, symbol):
        if self.vif.Rx.value.is_resolvable:
            return (0xFF & (symbol << 1)) | (0x1 & self.vif.Rx.value)
        return 0

    #   virtual protected def rxed(self,ref bit [7:0] chr):
    #   endfunction

uvm_component_utils(vip_monitor)



import unittest

from uvm.base.uvm_object import UVMObject
from uvm.base.uvm_factory import (
    UVMFactory, UVMDefaultFactory)
from uvm.comps.uvm_test import UVMTest
#from uvm.macros.uvm_factory import UVMFactory, UVMDefaultFactory
from uvm.macros import uvm_component_utils, uvm_object_utils
from uvm.base.uvm_coreservice import UVMCoreService


def createXXX():
    class XXX(UVMObject):
        pass
    uvm_object_utils(XXX)
    return XXX


def createYYY():
    class YYY(UVMObject):
        pass
    uvm_object_utils(YYY)
    return  YYY

def createLastOverride():
    class LastOverride(UVMObject):
        pass
    uvm_object_utils(LastOverride)
    return LastOverride


class TestUVMDefaultFactory(unittest.TestCase):

    def test_override(self):
        cs = UVMCoreService.get()
        fact = UVMDefaultFactory()
        cs.set_factory(fact)
        XXX = createXXX()
        YYY = createYYY()

        fact.set_type_override_by_name('XXX', 'YYY')
        ovrd = fact.find_override_by_type(requested_type=XXX.get_type(), full_inst_path='')
        self.assertIsNotNone(ovrd, 'Override XXX->YYY should exist')
        self.assertEqual(ovrd.get_type_name(), 'YYY')
        fact.debug_create_by_name('XXX')
        #fact.debug_create_by_type(XXX.get_type())


    def test_type_double_override(self):
        cs = UVMCoreService.get()
        fact = UVMDefaultFactory()
        cs.set_factory(fact)

        XXX = createXXX()
        YYY = createYYY()
        LastOverride = createLastOverride()

        fact.set_type_override_by_name('XXX', 'YYY')
        fact.set_type_override_by_name('XXX', 'LastOverride')
        self.assertFalse(fact.check_inst_override_exists(XXX.get_type(),
            LastOverride.get_type(), ''))
        ovrd = fact.find_override_by_type(requested_type=XXX.get_type(), full_inst_path='')
        self.assertIsNotNone(ovrd, 'Override XXX->LastOverride should exist')
        self.assertEqual(ovrd.get_type_name(), 'LastOverride')

        fact.set_type_override_by_name('YYY', 'LastOverride')
        ovrd = fact.find_override_by_type(requested_type=YYY.get_type(), full_inst_path='')
        self.assertIsNotNone(ovrd, 'Override YYY->LastOverride should exist')
        self.assertEqual(ovrd.get_type_name(), 'LastOverride')


    def test_inst_override(self):
        cs = UVMCoreService.get()
        fact = UVMDefaultFactory()
        cs.set_factory(fact)

        XXX = createXXX()
        LastOverride = createLastOverride()

        fact.set_inst_override_by_type(XXX.get_type(), LastOverride.get_type(),
                'top.my_inst.is_override')
        self.assertTrue(fact.check_inst_override_exists(XXX.get_type(),
            LastOverride.get_type(), 'top.my_inst.is_override'))
        self.assertFalse(fact.check_inst_override_exists(XXX.get_type(),
            LastOverride.get_type(), 'top.my_inst2'))

        obj = fact.create_object_by_type(XXX.get_type(), 'top.my_inst', 'is_override')
        obj2 = fact.create_object_by_type(XXX.get_type(), 'top.my_inst2', 'is_xxx')
        self.assertEqual(obj.get_name(), 'is_override')
        self.assertEqual(obj.get_type_name(), 'LastOverride')
        self.assertEqual(obj2.get_name(), 'is_xxx')
        self.assertEqual(obj2.get_type_name(), 'XXX')

    def test_create_component_by_name(self):
        cs = UVMCoreService.get()
        factory = UVMFactory.get()
        cs.set_factory(factory)

        class MyTest123(UVMTest):
            def __init__(self, name, parent):
                UVMTest.__init__(self, name, parent)

        uvm_component_utils(MyTest123)
        test_name = 'MyTest123'
        uvm_test_top = factory.create_component_by_name(test_name,
            "", "uvm_test_top", None)
        self.assertEqual(uvm_test_top.get_full_name(), "uvm_test_top")


    def test_create_object_by_name(self):
        factory = UVMFactory.get()

        class MyObj(UVMObject):
            def __init__(self, name):
                UVMObject.__init__(self, name)

        uvm_object_utils(MyObj)

        new_obj = factory.create_object_by_name('MyObj', "", "my_obj_name")
        self.assertEqual(new_obj.get_name(), 'my_obj_name')


    def test_register(self):
        fact = UVMDefaultFactory()
        class NotWorks(UVMObject):
            type_name = "xxx"

        obj = NotWorks('my_name')
        fact.register(obj)


if __name__ == '__main__':
    unittest.main()




from .uvm_mem import *
from .uvm_mem_mam import *
from .uvm_reg import *
from .uvm_reg_adapter import *
from .uvm_reg_backdoor import *
from .uvm_reg_block import *
from .uvm_reg_cbs import *
from .uvm_reg_field import *
from .uvm_reg_fifo import *
from .uvm_reg_file import *
from .uvm_reg_indirect import *
from .uvm_reg_item import *
from .uvm_reg_map import *
from .uvm_reg_model import *
from .uvm_reg_predictor import *
from .uvm_reg_sequence import *
from .uvm_vreg import *
from .uvm_vreg_field import *
from .sequences import *




from .uvm_printer import UVMTablePrinter, UVMTreePrinter
from .uvm_comparer import UVMComparer
from .uvm_packer import UVMPacker

uvm_default_table_printer = UVMTablePrinter()

uvm_default_tree_printer = UVMTreePrinter()

uvm_default_printer = uvm_default_table_printer

uvm_default_comparer = UVMComparer()

uvm_default_packer = UVMPacker()



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer, Combine
from cocotb.utils import get_sim_time

from uvm import *

NUM_SEQS = 1
NUM_LOOPS = 1

#UVMDebug.full_debug()


def to_string(X):
    return X.convert2string()


BUS_READ = 0
BUS_WRITE = 1

STATUS_OK = 0
STATUS_NOT_OK = 1


class bus_trans(UVMSequenceItem):

    def __init__(self, name=""):
        UVMSequenceItem.__init__(self, name)
        self.addr = 0
        self.data = 0
        self.op = 0
        self.rand('addr', range(0, 1 << 30))
        self.rand('data', range(0, (1 << 30) - 1))
        self.rand('op', [BUS_READ, BUS_WRITE])

    def do_copy(self, rhs):
        rhs_ = rhs
        if not sv.cast(rhs_, rhs):
            uvm_error("do_copy", "cast failed, check type compatability")
        UVMSequenceItem.do_copy(self, rhs)

        self.addr = rhs_.addr
        self.data = rhs_.data
        self.op = rhs_.op

    def do_compare(self, rhs, comparer):
        rhs_ = rhs  # bus_trans rhs_
        #if(!$cast(rhs_, rhs))
        #        `uvm_fatal("do_compare", "cast failed, check type compatability")
        return (self.op == rhs_.op) and (self.addr == rhs_.addr) and (self.data == rhs_.data)

    def convert2string(self):
        return sv.sformatf("op %s: addr=%0h, data=%0h",
            str(self.op), self.addr, self.data)

uvm_object_utils(bus_trans)


class bus_req(bus_trans):
    def __init__(self, name=""):
        UVMSequenceItem.__init__(self, name)
uvm_object_utils(bus_req)


class bus_rsp(bus_trans):

    def __init__(self, name=""):
        bus_trans.__init__(self, name)
        self.status = 0

    def do_copy(self, rhs):
        rhs_ = rhs
        #bus_rsp rhs_
        #if(!$cast(rhs_, rhs))
        #    uvm_fatal("do_copy", "cast failed, check type compatability")

        super.do_copy(rhs_)
        self.status = rhs_.status

    def convert2string(self):
        return sv.sformatf("op %s, status=%s",
                bus_trans.convert2string(self), str(self.status))


uvm_object_utils(bus_rsp)


class my_driver(UVMDriver):

    def __init__(self, name, parent):
        UVMDriver.__init__(self, name, parent)
        self.data_array = {}  # int data_array[511:0]
        self.nitems = 0
        self.recording_detail = UVM_HIGH


    async def run_phase(self, phase):
        req = None
        rsp = None

        while True:
            uvm_info("DRIVER", "my_driver getting the req now..",
                UVM_MEDIUM)
            qreq = []
            await self.seq_item_port.get(qreq)

            self.nitems += 1
            req = qreq[0]
            self.begin_tr(req)
            rsp = bus_rsp("bus_rsp")
            rsp.set_id_info(req)

            # Actually do the read or write here
            if req.op == BUS_READ:
                rsp.addr = req.addr  # [8:0]
                if rsp.addr in self.data_array:
                    rsp.data = self.data_array[rsp.addr]
                else:
                    rsp.data = 0xFFFFABCD
                uvm_info("sending", to_string(rsp), UVM_MEDIUM)
            else:
                self.data_array[req.addr] = req.data
                uvm_info("sending", to_string(req), UVM_MEDIUM)
            await self.seq_item_port.put(rsp)
            self.end_tr(req)


uvm_component_utils(my_driver)


class SoCLevelSeq(UVMSequence):

    def __init__(self, name=""):
        UVMSequence.__init__(self, name)
        self.seq_delay_ns = 10
        self.rand('seq_delay', range(10, 90))

    async def body(self):
        sys0_seq = SubSysLevelSeq('sys0_seq')
        # await sys0_seq.start(None, self)
        await uvm_do_on_with(self, sys0_seq, None,
            lambda num_blk_seq: num_blk_seq == 10,
            lambda blk_level_delay_ns: blk_level_delay_ns in [10, 20, 30],
        )
        await Timer(self.seq_delay_ns, "NS")
        sys2_seq = SubSysLevelSeq('sys2_seq')
        # We don't randomize the 2nd sequence
        await sys2_seq.start(None, self)


class SubSysLevelSeq(UVMSequence):
    def __init__(self, name=""):
        UVMSequence.__init__(self, name)
        self.num_blk_seq = 5
        self.rand('num_blk_seq', range(1, 11))
        self.blk_level_delay_ns = 25
        self.rand('blk_level_delay_ns', range(5, 50))

    async def body(self):
        for i in range(self.num_blk_seq):
            blk_seq = BlockLevelSeq('blk_level_seq')
            await blk_seq.start(None, self)
            await Timer(self.blk_level_delay_ns, "NS")

uvm_object_utils(SubSysLevelSeq)


class BlockLevelSeq(UVMSequence):
    def __init__(self, name=""):
        UVMSequence.__init__(self, name)

    async def body(self):
        seqr = None
        arr = []
        if not UVMConfigDb.get(UVMRoot.get(), "", "SEQR", arr):
            uvm_fatal("NO_SEQR", "BlockLevelSeq did not get sequencer")
        seqr = arr[0]
        bs = bus_trans('bus_trans')
        #await self.start_item(bs, sequencer=seqr)
        self.m_sequencer = seqr
        await uvm_do_with(self, bs)
        #    lambda addr: addr in [0],
        #    lambda data: 0xFFFFFFFF
        #)
        #await self.finish_item(bs)



class sequenceA(UVMSequence):

    #local static integer g_my_id = 1
    g_my_id = 1

    def __init__(self, name=""):
        UVMSequence.__init__(self, name)
        self.my_id = sequenceA.g_my_id
        sequenceA.g_my_id += 1


    async def body(self):
        req = None  # REQ  req
        rsp = None  # RSP  rsp
        uvm_info("sequenceA", "Starting sequence", UVM_MEDIUM)

        for i in range(NUM_LOOPS):
            #req = uvm_create(req, self.m_sequencer)
            uvm_info("sequenceA", "Starting item " + str(i), UVM_MEDIUM)
            req = bus_req("my_req")
            await Timer(10, "NS")

            req.addr = (self.my_id * NUM_LOOPS) + i
            req.data = self.my_id + i + 55
            req.op   = BUS_WRITE

            if self.m_sequencer is None:
                raise Exception("Null sequencer CRASH")

            # uvm_send(req)
            await self.start_item(req)
            await self.finish_item(req)
            uvm_info("sequenceA", "getting response1 for " + str(i), UVM_MEDIUM)
            rsp = []
            await self.get_response(rsp)
            uvm_info("sequenceA", "got response1 for " + str(i) + ": " +
                    rsp[0].convert2string(), UVM_MEDIUM)

            #uvm_create(req)
            req = bus_req("my_req")
            req.addr = (self.my_id * NUM_LOOPS) + i
            req.data = 0
            req.op   = BUS_READ

            #uvm_send(req)
            await self.start_item(req)
            await self.finish_item(req)
            uvm_info("sequenceA", "getting response2 for " + str(i), UVM_MEDIUM)
            rsp = []
            await self.get_response(rsp)

            rsp = rsp[0]
            if rsp.data != (self.my_id + i + 55):
                uvm_error("SequenceA", sv.sformatf("Error, addr: %0d, expected data: %0d, actual data: %0d",
                    req.addr, req.data, rsp.data))
        uvm_info("sequenceA", "Finishing sequence", UVM_MEDIUM)

uvm_object_utils(sequenceA)


class env(UVMEnv):

    def __init__(self, name, parent):
        UVMEnv.__init__(self, name, parent)
        self.drv = None
        self.sqr = None
        self.error = False

    def build_phase(self, phase):
        UVMEnv.build_phase(self, phase)
        self.sqr = UVMSequencer("sequence_controller", self)
        UVMConfigDb.set(UVMRoot.get(), "", "SEQR", self.sqr)
        # create and connect driver
        self.drv = my_driver("my_driver", self)
        uvm_info("ENV", "Connecting ports now here", UVM_MEDIUM)
        self.drv.seq_item_port.connect(self.sqr.seq_item_export)


    async def run_phase(self, phase):
        phase.raise_objection(self)
        soc_seq = SoCLevelSeq('soc_seq')
        await soc_seq.start(None)
        phase.drop_objection(self)

    def check_phase(self, phase):
        if self.drv.nitems == 0:
            self.error = True
            self.uvm_report_error('env', sv.sformatf("nitems: Exp %0d, Got %0d",
                NUM_SEQS * NUM_LOOPS, self.drv.nitems))


uvm_component_utils(env)


@cocotb.test()
async def test_module_top(dut):
    # UVMPhase.m_phase_trace = True
    e = env("env_name", parent=None)
    uvm_info("top","In top initial block", UVM_MEDIUM)
    await run_test(dut=dut)
    sim_time = sim_time = get_sim_time('ns')

    if sim_time == 0:
        raise Exception('Sim time has not progressed')
    else:
        if sim_time < 50.0:
            msg = "Exp: > 50, got: " + str(sim_time)
            raise Exception('Wrong sim time at the end: ' + msg)
    if e.error is True:
        raise Exception('Env had errors')



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer
from cocotb.utils import get_sim_time
from cocotb.clock import Clock

from uvm import UVMObject, UVMComponent, UVMDebug, UVM_HEX, uvm_fatal
from uvm.base.sv import sv
from uvm.base.uvm_globals import run_test
from uvm.base.uvm_config_db import UVMConfigDb, UVMConfigDbOptions
from uvm.base.uvm_coreservice import UVMCoreService
from uvm.base.uvm_global_vars import uvm_default_tree_printer

from uvm.macros import uvm_object_utils

#  //--------------------------------------------------------------------
#  // lower
#  //--------------------------------------------------------------------


class lower(UVMComponent):

    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)
        self.data = 0
        self.str = ""
        self.comp_id = -1

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(10, "NS")
        print("{}: {} HI".format(get_sim_time(), self.get_full_name()))
        await Timer(10, "NS")
        if self.comp_id == 2:
            uvm_fatal("COMP_ID_2", "run_phase() should be killed before fatal")
        phase.drop_objection(self)

    def get_type_name(self):
        return "lower"

    def build_phase(self, phase):
        val_arr = []
        if UVMConfigDb.get(self, "", "data", val_arr):
            self.data = val_arr[0]
        val_arr = []
        if UVMConfigDb.get(self, "", "str", val_arr):
            self.str = val_arr[0]

    def do_print(self, printer):
        printer.print_field("data", self.data, 32)
        printer.print_string("str", self.str)


#  //--------------------------------------------------------------------
#  // myunit
#  //--------------------------------------------------------------------
class myunit(UVMComponent):

    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)
        self.l1 = lower("l1", self)
        self.l1.comp_id = 1
        self.l2 = lower("l2", self)
        # TODO set this to cause fatal: self.l2.comp_id = 2
        UVMConfigDb.set(self, "l1", "str", "hi")
        UVMConfigDb.set(self, "*", "data", 0x100)
        self.l1.data = 0x30
        self.l2.data = 0x40
        self.a = [None] * 5
        for i in range(5):
            self.a[i] = i*i

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(10, "NS")
        #10 $display("%0t: %s HI", $time, get_full_name())
        phase.drop_objection(self)

    def get_type_name(self):
        return "myunit"

    def do_print(self, printer):
        printer.print_array_header("a", len(self.a))
        for i in range(len(self.a)):
            printer.print_field(sv.sformatf("a[%0d]", i), self.a[i], 32, UVM_HEX, "[")
        printer.print_array_footer()

#  // Factory registration
#
#  //--------------------------------------------------------------------
#  // lower_wrapper
#  //--------------------------------------------------------------------
#  class lower_wrapper extends uvm_object_wrapper
#
#    function uvm_component create_component(string name, uvm_component parent)
#      lower u
#      u = new(name, parent)
#      return u
#    endfunction
#
#    function string get_type_name()
#      return "lower"
#    endfunction
#
#    static function bit register_me(); uvm_coreservice_t cs_ = uvm_coreservice_t::get()
#
#      uvm_factory factory = cs_.get_factory()
#      lower_wrapper w; w = new
#      factory.register(w)
#      return 1
#    endfunction
#
#    static bit is_registered = register_me()


#  //--------------------------------------------------------------------
#  // myunit_wrapper
#  //--------------------------------------------------------------------
#  class myunit_wrapper extends uvm_object_wrapper
#
#    function uvm_component create_component(string name, uvm_component parent)
#      myunit u
#      u = new(name, parent)
#      return u
#    endfunction
#
#    function string get_type_name()
#      return "myunit"
#    endfunction
#
#    static function bit register_me(); uvm_coreservice_t cs_ = uvm_coreservice_t::get()
#
#      uvm_factory factory = cs_.get_factory()
#      myunit_wrapper w; w = new
#      factory.register(w)
#      return 1
#    endfunction
#
#    static bit is_registered = register_me()
#  endclass
#

#  //--------------------------------------------------------------------
#  // mydata
#  //--------------------------------------------------------------------

class mydata(UVMObject):
    #    `uvm_object_utils(mydata)
    #
    def __init__(self, name="mydata"):
        UVMObject.__init__(self, name)
uvm_object_utils(UVMObject)

#  //--------------------------------------------------------------------
#  // mydata_wrapper
#  //--------------------------------------------------------------------
#  class mydata_wrapper extends uvm_object_wrapper
#
#    function uvm_object create_object(string name="")
#      mydata u
#      u = new
#      if(name !="") u.set_name(name)
#      return u
#    endfunction
#
#    function string get_type_name()
#      return "myobject"
#    endfunction
#
#    static function bit register_me(); uvm_coreservice_t cs_ = uvm_coreservice_t::get()
#
#      uvm_factory factory = cs_.get_factory()
#      mydata_wrapper w; w = new
#      factory.register(w)
#      return 1
#    endfunction
#
#    static bit is_registered = register_me()
#  endclass
#`endif


#----------------------------------------------------------------------
# top
#----------------------------------------------------------------------


async def initial1():
    await run_test()


async def initial2(mu):
    await Timer(5, "NS")
    if mu.l1 is not None:
        print(str(mu.l1.get_name()))
        # TODO seems to be failing with async-await
        #mu.l1.kill()
    else:
        raise Exception("Oh no.")


@cocotb.test()
async def test_module_top(dut):
    mu = myunit("mu", None)
    bar = mydata()
    cs_ = UVMCoreService.get()
    factory = cs_.get_factory()

    UVMConfigDb.set(None, "mu.*", "data", 101)
    UVMConfigDb.set(None, "mu.*", "str", "hi")
    UVMConfigDb.set(None, "mu.l1", "data", 55)
    UVMConfigDb.set(None, "mu.*", "obj", bar)
    mu.print_config_settings("", None, 1)
    uvm_default_printer = uvm_default_tree_printer
    mu.print_obj()
    factory.print_factory(1)
    mu.print_obj()

    # tpoikela: Clock required by ghdl, otherwise scheduler throws error
    cocotb.fork(Clock(dut.clk, 10, "NS").start())

    # Fork 2 initial tasks
    task1 = cocotb.fork(initial1())
    task2 = cocotb.fork(initial2(mu))
    await sv.fork_join([task1, task2])



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer
from uvm import (UVMEnv, UVMConfigDb, sv, uvm_top, UVM_DEFAULT)
from uvm.macros import *

from classA import A
from classB import B


class my_env(UVMEnv):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.debug = 0
        self.inst1 = None
        self.inst2 = None


    def build_phase(self, phase):
        super().build_phase(phase)
        UVMConfigDb.set(self, "inst1.u2", "v", 5)
        UVMConfigDb.set(self, "inst2.u1", "v", 3)
        UVMConfigDb.set(self, "inst1.*", "s", 0x10)

        sv.display("%s: In Build: debug = %0d", self.get_full_name(), self.debug)

        self.inst1 = A("inst1", self)
        self.inst2 = B("inst2", self)

    def connect_phase(self, phase):
        if self.inst1.u2.v != 5:
            uvm_error("CONF_VAL_ERR", "Value should 5 but got " +
                    str(self.inst1.u2.v))
        if self.inst1.u1.my_conf_obj is None:
            uvm_error("CONF_VAL_ERR", "Obj should've got value but was None")
        if self.inst1.u2.my_conf_obj is not None:
            uvm_error("CONF_VAL_ERR", "Obj should've got None but was something else")
        if self.inst1.u1.tag != 'I am tagged':
            uvm_error("STR_VAL_ERR", "Expected 'I am tagged'. Got: " +
                self.inst1.u1.tag)

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        uvm_top.print_topology()
        await Timer(10, "NS")
        phase.drop_objection(self)


uvm_component_utils_begin(my_env)
uvm_field_int('debug', UVM_DEFAULT)
uvm_component_utils_end(my_env)
